import unittest

from pykit import config
from pykit import ututil
from pykit import zkutil

dd = ututil.dd


class TestZKConf(unittest.TestCase):

    def test_specified(self):

        c = zkutil.ZKConf(
            hosts='hosts',
            tx_dir='tx_dir/',
            record_dir='record_dir/',
            seq_dir='seq_dir/',
            lock_dir='lock_dir/',
            node_id='node_id',
            auth=('digest', 'a', 'b'),
            acl=(('foo', 'bar', 'cd'), ('xp', '123', 'cdrwa'))
        )

        self.assertEqual('hosts',              c.hosts())
        self.assertEqual('tx_dir/',            c.tx_dir())
        self.assertEqual('record_dir/',        c.record_dir())
        self.assertEqual('seq_dir/',           c.seq_dir())
        self.assertEqual('lock_dir/',          c.lock_dir())
        self.assertEqual('node_id',            c.node_id())
        self.assertEqual(('digest', 'a', 'b'), c.auth())

        self.assertEqual((('foo', 'bar', 'cd'),
                          ('xp', '123', 'cdrwa')),
                         c.acl())

        self.assertEqual('lock_dir/',                 c.lock())
        self.assertEqual('lock_dir/a',                c.lock('a'))
        self.assertEqual('record_dir/',               c.record())
        self.assertEqual('record_dir/a',              c.record('a'))
        self.assertEqual('tx_dir/alive/',             c.tx_alive())
        self.assertEqual('tx_dir/alive/a',            c.tx_alive('a'))
        self.assertEqual('tx_dir/alive/0000000001',   c.tx_alive(1))
        self.assertEqual('tx_dir/state/',             c.tx_state())
        self.assertEqual('tx_dir/state/a',            c.tx_state('a'))
        self.assertEqual('tx_dir/state/0000000001',   c.tx_state(1))
        self.assertEqual('tx_dir/journal/',           c.journal())
        self.assertEqual('tx_dir/journal/a',          c.journal('a'))
        self.assertEqual('tx_dir/journal/0000000001', c.journal(1))
        self.assertEqual('tx_dir/txidset',            c.txidset())
        self.assertEqual('tx_dir/txid_maker',         c.txid_maker())
        self.assertEqual('seq_dir/',                  c.seq())
        self.assertEqual('seq_dir/a',                 c.seq('a'))

        self.assertEqual(zkutil.make_kazoo_digest_acl((('foo', 'bar', 'cd'), ('xp', '123', 'cdrwa'))),
                         c.kazoo_digest_acl())
        self.assertEqual(('digest', 'a:b'), c.kazoo_auth())

    def test_default(self):
        old = (
            config.zk_hosts,
            config.zk_tx_dir,
            config.zk_record_dir,
            config.zk_lock_dir,
            config.zk_node_id,
            config.zk_auth,
            config.zk_acl,
        )

        config.zk_hosts = 'HOSTS'
        config.zk_tx_dir = 'TX_DIR/'
        config.zk_record_dir = 'RECORD_DIR/'
        config.zk_seq_dir = 'SEQ_DIR/'
        config.zk_lock_dir = 'LOCK_DIR/'
        config.zk_node_id = 'NODE_ID'
        config.zk_auth = ('DIGEST', 'A', 'B')
        config.zk_acl = (('FOO', 'BAR', 'CD'), ('XP', '123', 'CDRWA'))

        c = zkutil.ZKConf()

        self.assertEqual('HOSTS',              c.hosts())
        self.assertEqual('TX_DIR/',            c.tx_dir())
        self.assertEqual('RECORD_DIR/',        c.record_dir())
        self.assertEqual('SEQ_DIR/',           c.seq_dir())
        self.assertEqual('LOCK_DIR/',          c.lock_dir())
        self.assertEqual('NODE_ID',            c.node_id())
        self.assertEqual(('DIGEST', 'A', 'B'), c.auth())

        self.assertEqual((('FOO', 'BAR', 'CD'),
                          ('XP', '123', 'CDRWA')),
                         c.acl())

        self.assertEqual('LOCK_DIR/',                 c.lock())
        self.assertEqual('LOCK_DIR/a',                c.lock('a'))
        self.assertEqual('RECORD_DIR/',               c.record())
        self.assertEqual('RECORD_DIR/a',              c.record('a'))
        self.assertEqual('SEQ_DIR/',                  c.seq())
        self.assertEqual('SEQ_DIR/a',                 c.seq('a'))
        self.assertEqual('TX_DIR/alive/',             c.tx_alive())
        self.assertEqual('TX_DIR/alive/a',            c.tx_alive('a'))
        self.assertEqual('TX_DIR/alive/0000000001',   c.tx_alive(1))
        self.assertEqual('TX_DIR/state/',             c.tx_state())
        self.assertEqual('TX_DIR/state/a',            c.tx_state('a'))
        self.assertEqual('TX_DIR/state/0000000001',   c.tx_state(1))
        self.assertEqual('TX_DIR/journal/',           c.journal())
        self.assertEqual('TX_DIR/journal/a',          c.journal('a'))
        self.assertEqual('TX_DIR/journal/0000000001', c.journal(1))
        self.assertEqual('TX_DIR/txidset',            c.txidset())
        self.assertEqual('TX_DIR/txid_maker',         c.txid_maker())

        self.assertEqual(zkutil.make_kazoo_digest_acl((('FOO', 'BAR', 'CD'), ('XP', '123', 'CDRWA'))),
                         c.kazoo_digest_acl())
        self.assertEqual(('DIGEST', 'A:B'), c.kazoo_auth())

        (
            config.zk_hosts,
            config.zk_tx_dir,
            config.zk_record_dir,
            config.zk_lock_dir,
            config.zk_node_id,
            config.zk_auth,
            config.zk_acl,
        ) = old




import errno
import hashlib
import logging
import platform
import socket
import threading
import time

OS = platform.system()

PORT_N = 3
PORT_RANGE = (40000, 60000)

DEFAULT_SLEEP_TIME = 0.01  # sec

logger = logging.getLogger(__name__)


class PortlockError(Exception):
    pass


class PortlockTimeout(PortlockError):
    pass


class Portlock(object):

    def __init__(self, key, timeout=1, sleep_time=None):

        self.key = key
        self.addr = str_to_addr(key)
        self.timeout = timeout
        self.sleep_time = sleep_time or DEFAULT_SLEEP_TIME
        self.socks = [None] * PORT_N

        self.thread_lock = threading.RLock()

    def try_lock(self):

        self.thread_lock.acquire()

        try:
            self._lock()

            if self.has_locked():
                return True
            else:
                self.socks = [None] * PORT_N
                self.thread_lock.release()
                return False

        except Exception:
            self.thread_lock.release()
            raise

    def has_locked(self):

        if OS == 'Linux':
            return self.socks[0] is not None

        # other OS

        return len([x for x in self.socks
                    if x is not None]) > len(self.socks) / 2

    def acquire(self):

        t0 = time.time()

        while True:

            if self.try_lock():
                return

            now = time.time()
            left = t0 + self.timeout - now
            if left > 0:
                slp = min([self.sleep_time, left + 0.001])
                time.sleep(slp)
            else:
                raise PortlockTimeout(
                    'portlock timeout: ' + repr(self.key), self.key)

    def release(self):

        if not self.has_locked():
            return

        for sock in self.socks:
            if sock is not None:
                try:
                    sock.close()
                except Exception:
                    pass

        self.socks = [None] * PORT_N

        self.thread_lock.release()

    def _lock(self):

        if OS == 'Linux':
            so = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            try:
                addr = '\0/portlock/' + self.key
                so.bind(addr)
                self.socks[0] = so
                logger.debug('success to bind: {addr}'.format(addr=addr))
            except socket.error as e:
                if e.errno == errno.EADDRINUSE:
                    logger.debug('failure to bind: {addr}'.format(addr=addr))
                else:
                    raise

            return

        # other OS

        for i in range(len(self.socks)):

            addr = (self.addr[0], self.addr[1] + i)

            so = self._socket()

            try:
                so.bind(addr)
                self.socks[i] = so
                logger.debug('success to bind: {addr}'.format(addr=addr))
            except socket.error as e:
                if e.errno == errno.EADDRINUSE:
                    logger.debug('failure to bind: {addr}'.format(addr=addr))
                else:
                    raise

    def _socket(self):
        return socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    def __enter__(self):
        self.acquire()
        return self

    def __exit__(self, typ, value, traceback):
        self.release()

    def __del__(self):
        self.release()


def str_to_addr(x):

    # builtin hash() gives bad distribution with sequencial values.
    # re-hash it with 32 bit fibonacci hash.
    # And finally embed it into ip and port

    r = hashlib.sha1(str(x)).hexdigest()
    r = int(r, 16)
    p = r % (PORT_RANGE[1] - PORT_RANGE[0]) + PORT_RANGE[0]

    return ("127.0.0.1", p)


def test_collision():

    import resource
    resource.setrlimit(resource.RLIMIT_NOFILE, (102400, 102400))

    dd = {}
    ls = []
    for i in range(1 << 15):
        key = str(hashlib.sha1(str(i)).hexdigest())
        lck = key
        print 'lock is', i, lck
        l = Portlock(lck, timeout=8)
        r = l.try_lock()
        if not r:
            print 'collide', i, l.addr
            print l.socks

        dd[l.addr] = i
        ls.append(l)



"""
unittest utility
"""

import inspect
import logging
import os
import time
import unittest

_glb = {
    'unittest_logger': None,
}

debug_to_stdout = os.environ.get('UT_DEBUG') == '1'


# TODO make this configurable
# logging.basicConfig(level='INFO',
#                     format='[%(asctime)s,%(process)d-%(thread)d,%(filename)s,%(lineno)d,%(levelname)s] %(message)s',
#                     datefmt='%H:%M:%S'
#                     )

# logger = logging.getLogger('kazoo')
# logger.setLevel('INFO')


class Timer(object):

    def __init__(self):
        self.start = None
        self.end = None

    def __enter__(self):
        self.start = time.time()
        return self

    def __exit__(self, errtype, errval, _traceback):
        self.end = time.time()

    def spent(self):
        return (self.end or time.time()) - self.start


class ContextFilter(logging.Filter):

    """
    Add correct func, line number, line info to log record.

    To fix the issue that when test case function use dd() instead of
    logger.debug(), logging alwasy print context info of dd(), but not the
    caller test_xxx.
    """

    def filter(self, record):

        # skip this function
        stack = inspect.stack()[1:]

        for i, (frame, path, ln, func, line, xx) in enumerate(stack):

            if (frame.f_globals.get('__name__') == 'pykit.ututil'
                    and func == 'dd'):

                # this frame is dd(), find the caller
                _, path, ln, func, line, xx = stack[i + 1]

                record._fn = os.path.basename(path)
                record._ln = ln
                record._func = func
                return True

        record._fn = record.filename
        record._ln = record.lineno
        record._func = record.funcName
        return True


def _init():

    if _glb['unittest_logger'] is not None:
        return

    # test_logutil might require this module and logutil is still under test!
    try:
        from pykit import logutil
        logger = logutil.make_logger(
            '/tmp',
            log_name='unittest',
            level='DEBUG',
            fmt=('[%(asctime)s'
                 ' %(_fn)s:%(_ln)d'
                 ' %(levelname)s]'
                 ' %(message)s'
                 )
        )
        logger.addFilter(ContextFilter())

        _glb['unittest_logger'] = logger

    except Exception as e:
        print repr(e) + ' while init root logger'


def dd(*msg):
    """
    debug level logging in a test case function test_xx.

    dd() write log to stdout if unittest verbosity is 2.
    And dd always write log to log file in /tmp dir.
    """

    s = ' '.join([x.encode('utf-8') if isinstance(x, unicode) else str(x)
                    for x in msg])

    _init()

    l = _glb['unittest_logger']
    if l:
        l.debug(s)

    if not debug_to_stdout:
        return

    print s


def get_ut_verbosity():
    """
    Return the verbosity setting of the currently running unittest
    program, or 0 if none is running.
    """

    frame = _find_frame_by_self(unittest.TestProgram)
    if frame is None:
        return 0

    self = frame.f_locals.get('self')

    return self.verbosity


def get_case_logger():
    """
    Get a case specific logger.
    The logger name is: `<module>.<class>.<function>`,
    such as: `pykit.strutil.test.test_strutil.TestStrutil.test_format_line`

    It must be called inside a test_* function of unittest.TestCase, or no
    correct module/class/function name can be found.
    """

    frame = _find_frame_by_self(unittest.TestCase)

    self = frame.f_locals.get('self')

    module_name = frame.f_globals.get('__name__')
    class_name = self.__class__.__name__
    func_name = frame.f_code.co_name

    nm = module_name + '.' + class_name + '.' + func_name

    logger = logging.getLogger(nm)
    for f in logger.filters:
        if isinstance(f, ContextFilter):
            break
    else:
        logger.addFilter(ContextFilter())

    return logger


def _find_frame_by_self(clz):
    """
    Find the first frame on stack in which there is local variable 'self' of
    type clz.
    """

    frame = inspect.currentframe()

    while frame:
        self = frame.f_locals.get('self')
        if isinstance(self, clz):
            return frame

        frame = frame.f_back

    return None



#!/usr/bin/env python2
# coding: utf-8

import os
import shelve
import time
import unittest

from pykit import config
from pykit import fsutil
from pykit import threadutil
from pykit import ututil

dd = ututil.dd

rst_keys = ['read', 'write', 'ioutil']


class TestIostat(unittest.TestCase):

    def test_iostat(self):

        force_remove(config.iostat_stat_path)

        with ututil.Timer() as t:
            rst = fsutil.iostat('/dev/sda1')
            dd(rst)

            self.assertIsNotNone(rst)
            self.assertEqual(set(rst_keys),
                             set(rst.keys()))

            for k in rst_keys:
                self.assertGreaterEqual(rst[k], 0)

            self.assertGreaterEqual(t.spent(), 1.0)

        # read again

        with ututil.Timer() as t:
            rst = fsutil.iostat('/dev/sda1')
            dd(rst)

            self.assertIsNotNone(rst)
            self.assertEqual(set(rst_keys),
                             set(rst.keys()))

            for k in rst_keys:
                self.assertGreaterEqual(rst[k], 0)

            self.assertAlmostEqual(0.0, t.spent(), delta=0.1)

    def test_duration_is_negative(self):

        force_remove(config.iostat_stat_path)

        with ututil.Timer() as t:
            rst = fsutil.iostat('/dev/sda1')
            dd(rst)

            self.assertIsNotNone(rst)
            self.assertEqual(set(rst_keys),
                             set(rst.keys()))
            self.assertGreaterEqual(t.spent(), 1.0)

        d = shelve.open(config.iostat_stat_path)
        x = d['sda1']
        x['ts'] += 2
        d['sda1'] = x
        d.close()

        with ututil.Timer() as t:
            rst = fsutil.iostat('/dev/sda1')
            dd(rst)

            self.assertIsNotNone(rst)
            self.assertEqual(set(rst_keys),
                             set(rst.keys()))
            self.assertGreater(rst['write'], 0)
            self.assertGreaterEqual(t.spent(), 1.0)

    def test_path(self):

        rst = fsutil.iostat(path='/')
        dd(rst)

        self.assertIsNotNone(rst)
        self.assertEqual(set(rst_keys),
                         set(rst.keys()))

        for k in rst_keys:
            self.assertGreaterEqual(rst[k], 0)

    def test_ioutil_heavy_load(self):

        sess = {'running': True}

        def _write(i):
            while sess['running']:
                with open('/tmp/pykit-test-write-' + str(i), 'w') as f:
                    f.write('a' * 1024*1024*10)

        th1 = threadutil.start_daemon(_write, args=(1, ))
        th2 = threadutil.start_daemon(_write, args=(2, ))

        time.sleep(2)
        force_remove('/tmp/pykit-iostat')

        with ututil.Timer() as t:
            rst = fsutil.iostat('/dev/sda1')
            dd(rst)

            self.assertAlmostEqual(100, rst['ioutil'], delta=30)
            self.assertGreaterEqual(t.spent(), 1.0)

        sess['running'] = False
        th1.join()
        th2.join()

        force_remove('/tmp/pykit-test-write-1')
        force_remove('/tmp/pykit-test-write-2')

    def test_config(self):

        force_remove('/tmp/foo')

        # write to default place and then ensure config changes stat path

        rst = fsutil.iostat('/dev/sda1')

        # change stat path

        old = config.iostat_stat_path
        config.iostat_stat_path = '/tmp/foo'

        with ututil.Timer() as t:
            rst = fsutil.iostat('/dev/sda1')
            dd(rst)

            self.assertGreaterEqual(t.spent(), 1.0)

        # should be able to read something.
        fsutil.read_file(config.iostat_stat_path)

        force_remove(config.iostat_stat_path)
        config.iostat_stat_path = old


def force_remove(fn):

    try:
        os.rmdir(fn)
    except:
        pass

    try:
        os.unlink(fn)
    except:
        pass



import os
import sys

import daemonize
from pykit import proc

foo_fn = '/tmp/foo'
bar_fn = '/tmp/bar'
pidfn = '/tmp/test_daemonize.pid'


def write_file(fn, cont):
    try:
        with open(fn, 'w') as f:
            f.write(cont)
    except Exception as e:
        print repr(e) + ' while write_file:' + fn
        raise


def run():
    code, out, err = proc.shell_script('/usr/sbin/lsof -p ' + str(os.getpid()))
    write_file(foo_fn, repr(out))


if __name__ == '__main__':

    fd = open(bar_fn, 'w')
    op = sys.argv[1]
    sys.argv[1] = 'start'
    if op == 'close':
        daemonize.daemonize_cli(run, pidfn, close_fds=True)
    else:
        daemonize.daemonize_cli(run, pidfn, close_fds=False)



#!/usr/bin/env python2
# coding: utf-8

import sys

fn = '/tmp/foo'


def write_file(fn, cont):
    try:
        with open(fn, 'w') as f:
            f.write(cont)
    except Exception:
        raise


if __name__ == '__main__':
    args = sys.argv[1:]
    write_file(fn, ''.join(args))



from .fsutil import (
    FSUtilError,
    NotMountPoint,

    assert_mountpoint,
    calc_checksums,
    get_all_mountpoint,
    get_device,
    get_device_fs,
    get_disk_partitions,
    get_mountpoint,
    get_path_fs,
    get_path_inode_usage,
    get_path_usage,
    makedirs,
    get_sub_dirs,
    read_file,
    remove,
    write_file,
)

from .cat import (
    SEEK_END,
    SEEK_START,

    CatError,
    LockTimeout,
    NoData,
    NoSuchFile,
    Cat
)

from .iostat import (
    DeviceNotFound,

    iostat,
)

__all__ = [
    "FSUtilError",
    "NotMountPoint",

    "assert_mountpoint",
    "calc_checksums",
    "get_all_mountpoint",
    "get_device",
    "get_device_fs",
    "get_disk_partitions",
    "get_mountpoint",
    "get_path_fs",
    "get_path_inode_usage",
    "get_path_usage",
    "makedirs",
    "get_sub_dirs",
    "read_file",
    "remove",
    "write_file",


    "SEEK_END",
    "SEEK_START",

    "CatError",
    "LockTimeout",
    "NoData",
    "NoSuchFile",
    "Cat",

    "DeviceNotFound",

    "iostat",
]



import logging
import random
import threading
import time
import unittest

from pykit import jobq
from pykit import threadutil
from pykit import ututil

dd = ututil.dd


def add1(args):
    return args + 1


def multi2(args):
    return args * 2


def multi2_sleep(args):
    time.sleep(0.02)
    return args * 2


def discard_even(args):
    if args % 2 == 0:
        return jobq.EmptyRst
    else:
        return args


def sleep_5(args):
    time.sleep(5)
    return args


class TestProbe(unittest.TestCase):

    def _start_jobq_in_thread(self, n_items, n_worker, keep_order=False):

        def _sleep_1(args):
            time.sleep(0.1)
            return args

        def _nothing(args):
            return args

        probe = {}
        th = threading.Thread(target=lambda: jobq.run(range(n_items),
                                                      [(_sleep_1, n_worker),
                                                       _nothing],
                                                      probe=probe,
                                                      keep_order=keep_order,
                                                      ))
        th.daemon = True
        th.start()

        return th, probe

    def test_probe_single_thread(self):

        cases = (
                (0.05, 1, '_sleep_1 is working on 1st'),
                (0.1,  1, '_sleep_1 is working on 2nd'),
                (0.2,  0, 'all done'),
        )

        th, probe = self._start_jobq_in_thread(3, 1)

        for sleep_time, doing, case_mes in cases:

            time.sleep(sleep_time)
            stat = jobq.stat(probe)

            self.assertEqual(doing, stat['doing'], case_mes)

            # qsize() is not reliable. do not test the value of it.
            self.assertTrue(stat['workers'][0]['input']['size'] >= 0)
            self.assertTrue(stat['workers'][0]['input']['capa'] >= 0)

        # use the last stat

        workers = stat['workers']
        self.assertEqual(2, len(workers))

        w0 = workers[0]
        self.assertEqual(True, w0['name'].endswith(':_sleep_1'))
        self.assertEqual(0,    w0['input']['size'])
        self.assertEqual(True, w0['input']['capa'] > 0)

        th.join()

    def test_probe_3_thread(self):

        cases = (
                (0.05, 3, '_sleep_1 is working on 1st 3 items'),
                (0.1,  3, '_sleep_1 is working on 2nd 3 items'),
                (0.4,  0, 'all done'),
        )

        th, probe = self._start_jobq_in_thread(10, 3)

        for sleep_time, doing, case_mes in cases:

            time.sleep(sleep_time)
            stat = jobq.stat(probe)

            self.assertEqual(doing, stat['doing'], case_mes)

        # use the last stat

        workers = stat['workers']
        self.assertEqual(2, len(workers))

        th.join()

    def test_probe_3_thread_keep_order(self):

        cases = (
                (0.05, 3, '_sleep_1 is working on 1st 3 items'),
                (0.1,  3, '_sleep_1 is working on 2nd 3 items'),
                (0.4,  0, 'all done'),
        )

        th, probe = self._start_jobq_in_thread(10, 3, keep_order=True)

        for sleep_time, doing, case_mes in cases:

            time.sleep(sleep_time)
            stat = jobq.stat(probe)

            self.assertEqual(doing, stat['doing'], case_mes)

        # use the last stat

        workers = stat['workers']
        self.assertEqual(2, len(workers))

        th.join()


class TestDispatcher(unittest.TestCase):

    def test_dispatcher_job_manager(self):

        n_threads = 3
        n_numbers = 1000
        rst = {}
        ordered = []

        def _collect_by_tid(ii):

            tid = threading.current_thread().ident
            if tid not in rst:
                rst[tid] = []

            rst[tid].append(ii)
            return ii

        def _collect(ii):
            ordered.append(ii)

        jm = jobq.JobManager([
            (_collect_by_tid, n_threads, lambda x: x % n_threads),
            (_collect, 1),
        ])

        # In dispatcher mode it does not allow to set thread_num to prevent out
        # of order output
        self.assertRaises(jobq.JobWorkerError, jm.set_thread_num, _collect_by_tid, 10)

        for i in range(n_numbers):
            jm.put(i)

            st = jm.stat()
            dd(st)

            self.assertIn('dispatcher', st['workers'][0])
            dstat = st['workers'][0]['dispatcher']
            self.assertEqual(n_threads, len(dstat))
            for ds in dstat:
                self.assertIn('input', ds)
                self.assertIn('output', ds)

        jm.join()

        self.assertEqual(n_threads, len(rst.keys()))

        for arr in rst.values():
            m = arr[0] % n_threads
            for i in arr:
                # a thread receives args with the same mod by `n_threads`
                self.assertEqual(m, i % n_threads)

        # with dispatcher, output are ordered
        self.assertEqual([x for x in range(n_numbers)],
                         ordered)

    def test_dispatcher_run(self):

        n_threads = 3
        n_numbers = 1000
        rst = {}
        ordered = []

        def _collect_by_tid(ii):

            tid = threading.current_thread().ident
            if tid not in rst:
                rst[tid] = []

            rst[tid].append(ii)
            return ii

        def _collect(ii):
            ordered.append(ii)

        jobq.run(range(n_numbers), [
            (_collect_by_tid, n_threads, lambda x: x % n_threads),
            (_collect, 1),
        ])

        self.assertEqual(n_threads, len(rst.keys()))

        for arr in rst.values():
            m = arr[0] % n_threads
            for i in arr:
                # a thread receives args with the same mod by `n_threads`
                self.assertEqual(m, i % n_threads)

        # with dispatcher, output are ordered
        self.assertEqual([x for x in range(n_numbers)],
                         ordered)


class TestTimeout(unittest.TestCase):

    def test_timeout(self):

        def _sleep_1(args):
            sleep_got.append(args)
            time.sleep(0.1)
            return args

        def collect(args):
            rst.append(args)

        # collect quits before it get any item from its input queue

        rst = []
        sleep_got = []
        jobq.run(range(10), [_sleep_1, collect], timeout=0.05)
        time.sleep(0.2)
        self.assertEqual([0], sleep_got)
        self.assertEqual([], rst)

        rst = []
        sleep_got = []
        jobq.run(range(10), [_sleep_1, collect], timeout=0.15)
        time.sleep(0.2)
        self.assertEqual([0, 1], sleep_got)
        self.assertEqual([0], rst)


class TestJobManager(unittest.TestCase):

    def test_manager(self):

        rst = []

        def _sleep(args):
            time.sleep(0.1)
            return args

        def _collect(args):
            rst.append(args)

        jm = jobq.JobManager([_sleep, _collect])

        t0 = time.time()
        for i in range(3):
            jm.put(i)

        jm.join()

        t1 = time.time()

        self.assertEqual([0, 1, 2], rst)
        self.assertTrue(-0.05 < t1 - t0 - 0.3 < 0.05)

    def test_join_timeout(self):

        def _sleep(args):
            time.sleep(0.1)

        jm = jobq.JobManager([_sleep])

        t0 = time.time()
        for i in range(3):
            jm.put(i)

        jm.join(timeout=0.1)

        t1 = time.time()

        self.assertTrue(0.09 < t1 - t0 < 0.11)

    def test_stat(self):

        def _pass(args):
            return args

        jm = jobq.JobManager([_pass])

        for i in range(3):

            jm.put(i)

            time.sleep(0.01)
            st = jm.stat()

            # each element get in twice: _pass and _blackhole
            self.assertEqual((i + 1) * 2, st['in'])
            self.assertEqual((i + 1) * 2, st['out'])
            self.assertEqual(0, st['doing'])

        jm.join()

    def test_stat_on_builtin_method(self):

        rst = []
        jm = jobq.JobManager([rst.append])

        # stat() read attribute `func.__module__`.
        # But for builtin method, there is no __module__ attribute.

        # this should not raise
        jm.stat()

        jm.join()

    def test_set_thread_num(self):

        def _pass(args):
            return args

        rst = []

        jm = jobq.JobManager([_pass, rst.append])

        for invalid in (0, -1, 1.1):
            self.assertRaises(
                AssertionError, jm.set_thread_num, _pass, invalid)

        n = 10240
        for i in range(n):

            jm.put(i)

            # change thread number every 91 put
            if i % 91:
                # randomly change thread number
                jm.set_thread_num(_pass, i % 3 + 1)

        jm.join()

        rst.sort()
        for i in range(n):
            self.assertEqual(i, rst[i])

    def test_set_thread_num_with_object_method(self):
        """
        In python2, `x = X(); x.meth is x.meth` results in a `False`.
        Every time to retrieve a method, python creates a new **bound** function.

        See https://stackoverflow.com/questions/15977808/why-dont-methods-have-reference-equality
        """

        class X(object):

            def meth(self):
                pass

        x = X()

        jm = jobq.JobManager([x.meth])

        before = jm.stat()
        self.assertEqual(1, before['workers'][0]['nr_worker'])

        # This should not raise JobWorkerNotFound.
        jm.set_thread_num(x.meth, 2)

        after = jm.stat()
        self.assertEqual(2, after['workers'][0]['nr_worker'])

        jm.join()

    def test_set_thread_num_keep_order(self):

        def _pass(args):
            return args

        rst = []

        jm = jobq.JobManager([_pass, rst.append], keep_order=True)

        setter = {'running': True}

        def _change_thread_nr():
            while setter['running']:
                jm.set_thread_num(_pass, random.randint(1, 4))
                time.sleep(0.5)

        ths = []
        for ii in range(3):
            th = threadutil.start_daemon_thread(_change_thread_nr)
            ths.append(th)

        n = 10240
        for i in range(n):
            jm.put(i)

        jm.join()

        rst.sort()
        for i in range(n):
            self.assertEqual(i, rst[i])

        setter['running'] = False

        for th in ths:
            th.join()


class TestJobQ(unittest.TestCase):

    def test_exception(self):

        # Add a handler, or python complains "no handler assigned
        # to...."
        jl = logging.getLogger('pykit.jobq')
        jl.addHandler(logging.NullHandler())

        def err_on_even(args):
            if args % 2 == 0:
                raise Exception('even number')
            else:
                return args

        def collect(args):
            rst.append(args)

        rst = []
        jobq.run(range(10), [err_on_even, collect])
        self.assertEqual(list(range(1, 10, 2)), rst)

        # remove NullHandler
        jl.handlers = []

    def test_sequential(self):

        cases = (
                ([0, 1, 2], [add1], [1, 2, 3]),
                ([0, 1, 2], [add1, multi2], [2, 4, 6]),
                (range(3), [add1, (multi2, 1)], [2, 4, 6]),
                (range(100), [add1, (multi2, 1), discard_even], []),
                (range(100), [add1, discard_even], list(range(1, 101, 2))),
                (range(1024 * 10), [add1, multi2],
                 list(range(2, 1024 * 10 * 2 + 2, 2))),
        )

        def collect(args):
            rst.append(args)

        for inp, workers, out in cases:
            rst = []
            jobq.run(inp, workers + [collect])
            self.assertEqual(out, rst)

    def test_concurrent(self):

        cases = (
                (list(range(100)), [add1, (multi2_sleep, 10)],
                 list(range(2, 202, 2))
                 ),
                (range(100), [add1, (multi2_sleep, 10)],
                 list(range(2, 202, 2))
                 ),
                (range(1024 * 10), [add1, (multi2, 4)],
                 list(range(2, 1024 * 10 * 2 + 2, 2))
                 ),
        )

        def collect(args):
            rst.append(args)

        for inp, workers, out in cases:
            rst = []
            jobq.run(inp, workers + [collect])
            self.assertEqual(set(out), set(rst))

            rst = []
            jobq.run(inp, workers + [collect], keep_order=True)
            self.assertEqual(out, rst)

    def test_generator(self):

        def gen(args):
            for i in range(3):
                yield i
                time.sleep(0.1)

        def collect(args):
            rst.append(args)

        rst = []
        jobq.run(range(3), [(gen, 2), collect], keep_order=True)
        self.assertEqual([0, 1, 2] * 3, rst,
                         "generator should extract all before next")

        rst = []
        jobq.run(range(3), [(gen, 2), collect], keep_order=False)
        self.assertEqual(set([0, 1, 2]), set(rst),
                         "generator should get all")

        self.assertEqual(9, len(rst), 'nr of elts')


class TestDefaultTimeout(unittest.TestCase):

    def test_default_timeout_is_not_too_large(self):

        # Issue: threading.Thread.join does not accept a very large timeout
        # value

        def _sleep_1(args):
            time.sleep(0.02)

        jobq.run(range(1), [_sleep_1])


class TestLimitJobSpeed(unittest.TestCase):

    def test_limit_job_speed(self):

        job_num = 300
        job_speed = 100

        def entry_iter():
            for ii in xrange(job_num):
                yield ii

        def empty(num):
            pass

        t0 = time.time()

        jobq.run(entry_iter(), [
                (jobq.limit_job_speed(job_speed, 1), 1),
                (empty, 10),
        ])

        self.assertEqual(int(job_num / job_speed), int(time.time() - t0))



#! /usr/bin/env python2
# coding: utf-8

import unittest
import datetime

from pykit import utfyaml
from pykit import ututil

dd = ututil.dd


class TestUtfyaml(unittest.TestCase):

    def test_load_scalar(self):

        cases = (
            (
                'hello, 中国',
                {},
                'hello, 中国',

                'test load string',
            ),

            (
                'hello, 中国',
                {
                    'encoding': None,
                },
                u'hello, 中国',

                'test load string and output with unicode',
            ),

            (
                '123',
                {},
                123,

                'test load number',
            ),

            (
                '3.14',
                {},
                3.14,

                'test load float',
            ),

            (
                'True',
                {},
                True,

                'test load bool',
            ),

            (
                '2018-04-17t15:09:43.10+08:00',
                {},
                datetime.datetime(2018, 4, 17, 7, 9, 43, 100000),

                'test load datetime object',
            ),

            (
                "!!python/unicode 'hello'",
                {},
                u'hello',

                'test load tag',
            ),
        )

        for src, kwargs, expected, msg in cases:

            self._test_yaml(utfyaml.load, src, expected, msg, **kwargs)


    def test_load_list(self):

        cases = (
            (
                '[猫, mouse, 狗]',
                {},
                ['猫', 'mouse', '狗'],

                'test normal load list',
            ),

            (
                '[猫, mouse, 狗]',
                {
                    'encoding': None,
                },
                [u'猫', u'mouse', u'狗'],

                'test normal load list and load str as unicode',
            ),

            (
                (
                    '- 猫\n'
                    '- mouse\n'
                    '- 狗\n'
                    '- goldfish\n'
                ),
                {},
                ['猫', 'mouse', '狗', 'goldfish'],

                'test load list with "-" style',
            ),

            (
                (
                    '- 猫\n'
                    '- mouse\n'
                    '- 狗\n'
                    '- goldfish\n'
                ),
                {
                    'encoding': None,
                },
                [u'猫', u'mouse', u'狗', u'goldfish'],

                'test load list with "-" style and load str as unicode',
            ),

            (
                '[1, two, 三, true, 3.14]',
                {},
                [1, 'two', '三', True, 3.14],

                'test load mixed'
            ),

            (
                '[1, two, 三]',
                {
                    'encoding': None,
                },
                [1, u'two', u'三'],

                'test load mixed and load str as unicode',
            ),

            (
                "[!!python/unicode 'hello', !!python/tuple ['tuple', 'value']]",
                {},
                [u'hello', ('tuple', 'value')],

                'test load list with tag',
            ),
        )

        for src, kwargs, expected, msg in cases:

            self._test_yaml(utfyaml.load, src, expected, msg, **kwargs)


    def test_load_dict(self):

        cases = (
            (
                (
                    'number: 1\n'
                    'string: hello\n'
                    'bool: [True, False]\n'
                    '汉字: 我\n'
                    "!!python/unicode 'unicode': !!python/unicode 'hello'\n"
                    "!!python/tuple [tuple, key]: !!python/tuple [tuple, value]\n"
                ),
                {},
                {
                    'number': 1,
                    'string': 'hello',
                    'bool': [True, False],
                    '汉字': '我',
                    u'unicode': u'hello',
                    ('tuple', 'key'): ('tuple', 'value'),
                },

                'test normal load dict',
            ),

            (
                (
                    'number: 1\n'
                    'string: hello\n'
                    'bool: [True, False]\n'
                    '汉字: 我\n'
                ),
                {
                    'encoding': None,
                },
                {
                    u'number': 1,
                    u'string': u'hello',
                    u'bool': [True, False],
                    u'汉字': u'我',
                },

                'test load dict and load str as unicode'
            ),

            (
                (
                    u'中国:\n'
                    u'  北京:\n'
                    u'    - 朝阳\n'
                    u'    - 海淀\n'
                ),
                {},
                {
                    '中国': {
                        '北京': ['朝阳', '海淀'],
                    },
                },

                'test load dict with indentation'
            ),

            (
                (
                    u'中国:\n'
                    u'  北京:\n'
                    u'    - 朝阳\n'
                    u'    - 海淀\n'
                ),
                {
                    'encoding': None,
                },
                {
                    u'中国': {
                        u'北京': [u'朝阳', u'海淀'],
                    },
                },

                'test load dict with indentation and load str as unicode'
            ),

            (
                '{hash: { name: Steve, foo: bar}}',
                {},
                {
                    'hash': {
                        'name': 'Steve',
                        'foo': 'bar',
                    },
                },

                'test load dict with "{}" style'
            ),

            (
                '{hash: { name: Steve, foo: bar}}',
                {
                    'encoding': None,
                },
                {
                    u'hash': {
                        u'name': u'Steve',
                        u'foo': u'bar',
                    },
                },

                'test load dict with "{}" style and load str as unicode'
            ),
        )

        for src, kwargs, expected, msg in cases:

            self._test_yaml(utfyaml.load, src, expected, msg, **kwargs)


    def test_dump_scalar(self):

        cases = (
            (
                'hello, 中国',
                {},
                'hello, 中国\n',

                'test normal dump scalar',
            ),

            (
                'hello, 中国',
                {
                    'encoding': None,
                },
                u'hello, 中国\n',

                'test output with unicode',
            ),

            (
                'hello, 中国',
                {
                    'encoding': 'GBK',
                },
                'hello, \xd6\xd0\xb9\xfa\n',

                'test output with GBK',
            ),

            (
                u'hello, 中国',
                {
                    'save_unicode': True,
                },
                "!!python/unicode 'hello, 中国'\n",

                'test output with utf-8 and save unicode as a object',
            ),

            (
                u'hello, 中国',
                {
                    'encoding': None,
                    'save_unicode': True,
                },
                u"!!python/unicode 'hello, 中国'\n",

                'test output with unicode and save unicode as a object',
            ),

            (
                u'hello, 中国',
                {
                    'encoding': 'GBK',
                    'save_unicode': True,
                },
                "!!python/unicode 'hello, \xd6\xd0\xb9\xfa'\n",

                'test output with unicode and save unicode as a object',
            ),

            (
                True,
                {},
                'true\n',

                'test dump bool',
            ),

            (
                123,
                {},
                '123\n',

                'test dump number',
            ),

            (
                3.14,
                {},
                '3.14\n',

                'test dump float',
            ),

            (
                datetime.datetime(2018, 7, 17),
                {},
                '2018-07-17 00:00:00\n',

                'test dump datetime object'
            ),
        )

        for src, kwargs, expected, msg in cases:

            self._test_yaml(utfyaml.dump, src, expected, msg, **kwargs)


    def test_dump_list(self):

        cases = (

            (
                ['猫', 'mouse', '狗'],
                {},
                '[猫, mouse, 狗]\n',

                'test normal dump list',
            ),

            (
                ['猫', 'mouse', '狗'],
                {
                    'encoding': None,
                },
                u'[猫, mouse, 狗]\n',

                'test output with unicode',
            ),

            (
                ['猫', 'mouse', '狗'],
                {
                    'encoding': 'GBK',
                },
                '[\xc3\xa8, mouse, \xb9\xb7]\n',

                'test output with GBK',
            ),

            (
                [u'猫', u'and', u'老鼠'],
                {
                    'save_unicode': True,
                },
                "[!!python/unicode '猫', !!python/unicode 'and', !!python/unicode '老鼠']\n",

                'test output with utf-8 and save unicode as object',
            ),

            (
                [u'猫', u'mouse', u'狗'],
                {
                    'encoding': 'GBK',
                    'save_unicode': True,
                },
                "[!!python/unicode '\xc3\xa8', !!python/unicode 'mouse', !!python/unicode '\xb9\xb7']\n",

                'test output with GBK and save unicode as object',
            ),
        )

        for src, kwargs, expected, msg in cases:

            self._test_yaml(utfyaml.dump, src, expected, msg, **kwargs)


    def test_dump_dict(self):

        src = {
            0: 'number key',

            'China': {
                'Captial': '北京',
                'Language': '普通话',
            },

            'bool': True,
            'number': 1,
            'string': 'hello',
            u'unicode': u'hello, 中文',
            '汉字': '我',

            ('tuple', 'key'): ('tuple', 'value'),
        }

        cases = (
            (
                {},
                (
                    '0: number key\n'
                    'China: {Captial: 北京, Language: 普通话}\n'
                    'string: hello\n'
                    'unicode: hello, 中文\n'
                    '汉字: 我\n'
                    'bool: true\n'
                    '? !!python/tuple [tuple, key]\n'
                    ': !!python/tuple [tuple, value]\n'
                    'number: 1\n'
                ),

                'normal test dump dict'
            ),

            (
                {
                    'encoding': None,
                },
                (
                    u'0: number key\n'
                    u'China: {Captial: 北京, Language: 普通话}\n'
                    u'string: hello\n'
                    u'unicode: hello, 中文\n'
                    u'汉字: 我\n'
                    u'bool: true\n'
                    u'? !!python/tuple [tuple, key]\n'
                    u': !!python/tuple [tuple, value]\n'
                    u'number: 1\n'
                ),

                'test output with unicode'
            ),

            (
                {
                    'encoding': 'GBK',
                },
                (
                    '0: number key\n'
                    'China: {Captial: \xb1\xb1\xbe\xa9, Language: \xc6\xd5\xcd\xa8\xbb\xb0}\n'
                    'string: hello\n'
                    'unicode: hello, \xd6\xd0\xce\xc4\n'
                    '\xba\xba\xd7\xd6: \xce\xd2\n'
                    'bool: true\n'
                    '? !!python/tuple [tuple, key]\n'
                    ': !!python/tuple [tuple, value]\n'
                    'number: 1\n'
                ),

                'test output with GBK'
            ),

            (
                {
                    'save_unicode': True,
                },
                (
                    '0: number key\n'
                    'China: {Captial: 北京, Language: 普通话}\n'
                    'string: hello\n'
                    "!!python/unicode 'unicode': !!python/unicode 'hello, 中文'\n"
                    '汉字: 我\n'
                    'bool: true\n'
                    '? !!python/tuple [tuple, key]\n'
                    ': !!python/tuple [tuple, value]\n'
                    'number: 1\n'
                ),

                'test output with utf-8 and save unicode as a object'
            ),

            (
                {
                    'encoding': 'GBK',
                    'save_unicode': True,
                },
                (
                    '0: number key\n'
                    'China: {Captial: \xb1\xb1\xbe\xa9, Language: \xc6\xd5\xcd\xa8\xbb\xb0}\n'
                    'string: hello\n'
                    "!!python/unicode 'unicode': !!python/unicode 'hello, \xd6\xd0\xce\xc4'\n"
                    '\xba\xba\xd7\xd6: \xce\xd2\n'
                    'bool: true\n'
                    '? !!python/tuple [tuple, key]\n'
                    ': !!python/tuple [tuple, value]\n'
                    'number: 1\n'
                ),

                'test output with GBK and save unicode as a object'
            ),
        )

        for kwargs, expected, msg in cases:

            self._test_yaml(utfyaml.dump, src, expected, msg, **kwargs)


    def _test_yaml(self, _exec, src, expected, msg, **kwargs):

            dd('msg: ', msg)
            dd('expected: ', expected)

            rst = _exec(src, **kwargs)

            dd('result  : ', rst)

            self.assertEqual(expected, rst)



import os
import time
import unittest

from pykit import proc
from pykit import ututil

dd = ututil.dd

this_base = os.path.dirname(__file__)


class TestProcError(unittest.TestCase):

    foo_fn = '/tmp/foo'

    def _read_file(self, fn):
        try:
            with open(fn, 'r') as f:
                cont = f.read()
                return cont
        except EnvironmentError:
            return None

    def _clean(self):

        # remove written file
        try:
            os.unlink(self.foo_fn)
        except EnvironmentError:
            pass

    def setUp(self):
        self._clean()

    def tearDown(self):
        self._clean()

    def test_procerror(self):
        ex_args = (1, 'out', 'err', 'ls', ('a', 'b'), {"close_fds": True})
        ex = proc.ProcError(*ex_args)

        self.assertEqual(ex_args, (ex.returncode,
                                   ex.out,
                                   ex.err,
                                   ex.command,
                                   ex.arguments,
                                   ex.options))

        self.assertEqual(ex_args, ex.args)

    def test_code_out_err(self):

        subproc = os.path.join(this_base, 'subproc.py')

        returncode, out, err = proc.command('python2', subproc, '222')

        self.assertEqual(222, returncode)
        self.assertEqual('out-1\nout-2\n', out)
        self.assertEqual('err-1\nerr-2\n', err)

        try:
            returncode, out, err = proc.command_ex('python2', subproc, '222')
        except proc.ProcError as e:
            self.assertEqual(222, e.returncode)
            self.assertEqual('out-1\nout-2\n', e.out)
            self.assertEqual('err-1\nerr-2\n', e.err)
            self.assertEqual('python2', e.command)
            self.assertTrue(e.arguments[0].endswith('subproc.py'))
            self.assertEqual('222', e.arguments[1])
            self.assertEqual({}, e.options)
        else:
            self.fail('expect proc.ProcError to be raised')

        returncode, out, err = proc.command_ex('python2', subproc, '0')
        self.assertEqual(0, returncode)
        self.assertEqual('out-1\nout-2\n', out)
        self.assertEqual('err-1\nerr-2\n', err)

        returncode, out, err = proc.command('python2', subproc, '0')

        self.assertEqual(0, returncode)
        self.assertEqual('out-1\nout-2\n', out)
        self.assertEqual('err-1\nerr-2\n', err)

    def test_close_fds(self):

        read_fd = os.path.join(this_base, 'read_fd.py')

        with open(read_fd) as f:
            fd = f.fileno()

            returncode, out, err = proc.command(
                'python2', read_fd, str(fd), close_fds=False)

            self.assertEqual(0, returncode)
            self.assertEqual('###\n', out)
            self.assertEqual('', err)

            returncode, out, err = proc.command(
                'python2', read_fd, str(fd), close_fds=True)

            self.assertEqual(1, returncode)
            self.assertEqual('errno=9\n', out)
            self.assertEqual('', err)

    def test_cwd(self):

        returncode, out, err = proc.command(
            'python2', 'subproc.py', '111', cwd=this_base)
        self.assertEqual(111, returncode)

        returncode, out, err = proc.command('python2', 'subproc.py', '111')
        # can not find subproc.py
        self.assertEqual(2, returncode)

    def test_env(self):
        returncode, out, err = proc.command('python2', 'print_env.py', 'abc',
                                            env={"abc": "xyz"},
                                            cwd=this_base)
        dd('returncode:', returncode)
        dd('out:', out)
        dd('err:', err)

        self.assertEqual(0, returncode)
        self.assertEqual('xyz\n', out)

    def test_stdin(self):

        returncode, out, err = proc.command('python2', 'read_fd.py', '0',
                                            stdin='abc',
                                            cwd=this_base)
        dd('returncode:', returncode)
        dd('out:', out)
        dd('err:', err)

        self.assertEqual(0, returncode)
        self.assertEqual('abc\n', out)

    def test_shell_script(self):

        returncode, out, err = proc.shell_script(
            'ls ' + this_base + ' | grep init | grep -v pyc')

        dd('returncode:', returncode)
        dd('out:', out)
        dd('err:', err)

        self.assertEqual(0, returncode)
        self.assertEqual('__init__.py\n', out)

    def test_start_process(self):

        cases = (
            ('python2', this_base + '/write.py', ['foo'], 'foo'),
            ('python2', this_base + '/write.py', ['foo', 'bar'], 'foobar'),
            ('sh', this_base + '/write.sh', ['123'], '123'),
            ('sh', this_base + '/write.sh', ['123', '456'], '123456'),
        )

        for cmd, target, args, expected in cases:
            proc.start_process(cmd, target, os.environ, *args)
            time.sleep(0.1)
            self.assertEqual(expected, self._read_file(self.foo_fn))



import logging
import os
import time

import redis
from kazoo.client import KazooClient

from pykit import threadutil
from pykit import utfjson
from pykit.cgrouparch import account
from pykit.cgrouparch import cgroup_manager
from pykit.cgrouparch import communicate

logger = logging.getLogger(__name__)


global_value = {}


def init_redis_client(context):
    client = redis.StrictRedis(context['redis_ip'], context['redis_port'])
    context['redis_client'] = client


def get_zk_client(context):
    host = context['get_zk_host']()
    zk_client = KazooClient(hosts=host, timeout=3.0, randomize_hosts=True,
                            auth_data=context['zk_auth_data'], logger=logger)

    zk_client.start()

    return zk_client


def update_conf(event):
    logger.info('update conf triggered at: %f' % time.time())

    context = global_value['context']

    zk_path = '%s/arch_conf' % context['zk_prefix']

    while True:
        try:
            zk_client = context['zk_client']
            resp = zk_client.get(zk_path, watch=update_conf)
            break

        except Exception as e:
            logger.exception('failed to get from zk: ' + repr(e))
            time.sleep(5)

    context['arch_conf'] = {
        'version': resp[1].version,
        'value': utfjson.load(resp[0]),
    }

    logger.info('arch conf in zk changed at: %f, current verrsion: %d' %
                (time.time(), resp[1].version))

    cgroup_manager.build_all_subsystem_cgroup_arch(context)


def on_lost(stat):
    logger.warn('zk client on lost, stat is: %s, about to exit' % str(stat))
    os._exit(2)


def init_arch_conf(context):
    while True:
        try:
            if context['zk_client'] is None:
                context['zk_client'] = get_zk_client(context)
                context['zk_client'].add_listener(on_lost)

            zk_path = '%s/arch_conf' % context['zk_prefix']
            resp = context['zk_client'].get(zk_path, watch=update_conf)

            context['arch_conf'] = {
                'version': resp[1].version,
                'value': utfjson.load(resp[0]),
            }

            return

        except Exception as e:
            logger.warn('failed to get arch conf from zk: %s' % repr(e))

            try:
                context['zk_client'].stop()
            except Exception as e:
                logger.info('failed to stop zk client: ' + repr(e))

            context['zk_client'] = None
            time.sleep(10)


def update_cgexec_arg(cgexec_arg, subsystem_name, cgroup_relative_path,
                      cgroup_conf):
    sub_cgroup = cgroup_conf.get('sub_cgroup')
    if sub_cgroup is None:
        return

    for sub_cgroup_name, sub_cgroup_conf in sub_cgroup.iteritems():
        sub_relative_path = cgroup_relative_path + '/' + sub_cgroup_name

        if sub_cgroup_name in cgexec_arg:
            cgexec_arg[sub_cgroup_name] += ' -g %s:%s' % (
                subsystem_name, sub_relative_path)

        update_cgexec_arg(cgexec_arg, subsystem_name, sub_relative_path,
                          sub_cgroup_conf)


def get_cgexec_arg(cgroup_names, **argkv):
    context = {
        'cgroup_dir': argkv.get('cgroup_dir', '/sys/fs/cgroup'),
        'get_zk_host': argkv['get_zk_host'],
        'zk_prefix': argkv['zk_prefix'],
        'zk_auth_data': argkv['zk_auth_data'],
    }

    cgexec_arg = {}
    for cgroup_name in cgroup_names:
        cgexec_arg[cgroup_name] = ''

    try:
        zk_client = get_zk_client(context)

        zk_path = '%s/arch_conf' % context['zk_prefix']
        resp = zk_client.get(zk_path)

        zk_client.stop()

        context['arch_conf'] = {
            'version': resp[1].version,
            'value': utfjson.load(resp[0]),
        }
        cgroup_manager.build_all_subsystem_cgroup_arch(context)

        arch_conf_value = context['arch_conf']['value']
        for subsystem_name, subsystem_conf in arch_conf_value.iteritems():
            cgroup_relative_path = ''
            cgroup_conf = subsystem_conf
            update_cgexec_arg(cgexec_arg, subsystem_name,
                              cgroup_relative_path, cgroup_conf)

        return cgexec_arg

    except Exception as e:
        logger.exception('failed to get cgexec arg: ' + repr(e))
        return cgexec_arg


def run(**argkv):
    context = {
        'get_cgroup_pid_file': argkv['get_cgroup_pid_file'],
        'cgroup_dir': argkv.get('cgroup_dir', '/sys/fs/cgroup'),

        'communicate_ip': argkv.get('communicate_ip', '0.0.0.0'),
        'communicate_port': argkv.get('communicate_port', 43409),

        'tasks_update_interval': argkv.get('tasks_update_interval', 30),

        'redis_ip': argkv['redis_ip'],
        'redis_port': argkv['redis_port'],
        'redis_prefix': argkv.get('redis_prefix', 'cgroup_arch'),
        'redis_client': None,
        'redis_expire_time': argkv.get('redis_expire_time', 60 * 5),

        'get_zk_host': argkv['get_zk_host'],
        'zk_prefix': argkv['zk_prefix'],
        'zk_auth_data': argkv['zk_auth_data'],
        'zk_client': None,

        'protected_cgroup': argkv.get('protected_cgroup'),

        'arch_conf': None,
    }

    init_redis_client(context)
    init_arch_conf(context)

    global_value['context'] = context

    cgroup_manager.build_all_subsystem_cgroup_arch(context)

    cgroup_manager.set_cgroup(context)

    cgroup_manager.reset_statistics(context)

    threadutil.start_daemon_thread(account.run, args=(context,))

    threadutil.start_daemon_thread(cgroup_manager.loop_set_cgroup,
                                   args=(context,))

    communicate.run(context, ip=context['communicate_ip'],
                    port=context['communicate_port'])



import threading
import time
import unittest

from pykit import portlock
from pykit import ututil

dd = ututil.dd


class TestPortlock(unittest.TestCase):

    def test_mem_test(self):

        a = portlock.Portlock('k1')
        b = portlock.Portlock('k1')

        dd('lock-a: ', a.addr)
        dd('lock-b: ', b.addr)

        r = b.has_locked()
        self.assertFalse(r, 'not locked')

        r = b.has_locked()
        self.assertFalse(r, 'still not locked')

        a.acquire()

        r = b.has_locked()
        self.assertFalse(r, 'not locked by me.')

        r = b.try_lock()
        self.assertFalse(r, 'can not lock while another one hold it')

        b.release()
        r = b.try_lock()
        self.assertFalse(r, 'b can not release lock belongs to a')

        a.release()

        r = b.try_lock()
        self.assertTrue(r, 'I can hold it now')
        dd('try-locked:', r)

        try:
            a.acquire()
            self.fail("lock-a should not be able to get lock")
        except portlock.PortlockError:
            pass

    def test_1_lock_per_thread(self):

        sess = {'n': 0}

        def worker(l, ident):
            for ii in range(1000):
                l.acquire()
                dd('{0}-{1} start'.format(ident, ii))

                self.assertEqual(
                    0, sess['n'], 'n is 0 just after lock is acquired, 1-lock-for-1')

                sess['n'] += 1
                time.sleep(0.001)
                self.assertEqual(
                    1, sess['n'], "no more than 2 thread holding lock")
                sess['n'] -= 1
                dd('{0}-{1} end'.format(ident, ii))
                l.release()

        ts = [threading.Thread(
            target=worker,
            args=(portlock.Portlock('x', timeout=100),
                  x))
              for x in range(10)]

        for t in ts:
            t.start()

        for t in ts:
            t.join()

    def test_1_lock_for_all_thread(self):

        sess = {'n': 0}

        def worker(l):
            for ii in range(1000):
                l.acquire()

                self.assertEqual(
                    0, sess['n'], 'n is 0 just after lock is acquired, 1-lock-for-all')

                sess['n'] += 1
                time.sleep(0.001)
                self.assertEqual(
                    1,  sess['n'], "no more than 2 thread holding lock")
                sess['n'] -= 1
                l.release()

        lock = portlock.Portlock('x', timeout=100)
        ts = [threading.Thread(target=worker,
                               args=(lock, ))
              for x in range(3)]

        for t in ts:
            t.daemon = True
            t.start()

        for t in ts:
            t.join()

    def test_sleep_time(self):

        lock0 = portlock.Portlock('x')
        lock = portlock.Portlock('x', timeout=1, sleep_time=2)

        with lock0:

            t0 = time.time()
            try:
                lock.acquire()
            except portlock.PortlockTimeout:
                pass

            t1 = time.time()

            self.assertTrue(0.9 < t1 - t0 < 1.1,
                            'sleep_time should not affect timeout(1 sec),'
                            ' but it spends {0} seconds'.format(t1 - t0))



#!/usr/bin/env python2
# coding: utf-8


import re
import string
import sys
import types


listtype = (types.TupleType, types.ListType)


def _findquote(line, quote):
    if len(quote) == 0:
        return -1, -1, []

    i = 0
    n = len(line)
    escape = []
    while i < n:
        if line[i] == '\\':
            escape.append(i)
            i += 2
            continue

        if line[i] in quote:
            quote_s = i - len(escape)

            j = i
            i += 1
            while i < n and line[i] != line[j]:
                if line[i] == '\\':
                    escape.append(i)
                    i += 2
                    continue

                i += 1

            if i < n:
                quote_e = i - len(escape)
                return quote_s, quote_e, escape
            else:
                return quote_s, -1, escape

        i += 1

    return -1, -1, escape


def tokenize(line, sep=None, quote='"\'', preserve=False):
    if sep == quote:
        raise ValueError, 'diffrent sep and quote is required'

    if sep is None:
        if len(line) == 0:
            return []
        line = line.strip()

    rst = ['']
    n = len(line)
    i = 0
    while i < n:
        quote_s, quote_e, escape = _findquote(line[i:], quote)

        if len(escape) > 0:
            lines = []
            x = 0
            for e in escape:
                lines.append(line[x:i + e])
                x = i + e + 1
            lines.append(line[x:])
            line = ''.join(lines)
            n = len(line)

        if quote_s < 0:
            sub = n
        else:
            sub = i + quote_s

        if i < sub:
            sub_rst = line[i:sub].split(sep)
            if sep is None:
                if line[sub - 1] in string.whitespace:
                    sub_rst.append('')
                if line[i] in string.whitespace:
                    sub_rst.insert(0, '')

            head = rst.pop()
            sub_rst[0] = head + sub_rst[0]
            rst += sub_rst

        if quote_s < 0:
            break

        # discard incomplete
        # 'a b"c'  ->  ['a']
        if quote_e < 0:
            rst.pop()
            break

        head = rst.pop()

        if preserve:
            head += line[i + quote_s:i + quote_e + 1]
        else:
            head += line[i + quote_s + 1:i + quote_e]

        rst.append(head)
        i += quote_e + 1

    return rst


def line_pad(linestr, padding=''):

    lines = linestr.split("\n")

    if type(padding) in types.StringTypes:
        lines = [padding + x for x in lines]

    elif callable(padding):
        lines = [padding(x) + x for x in lines]

    lines = "\n".join(lines)

    return lines


def format_line(items, sep=' ', aligns=''):
    '''
    format a line with multi-row columns.

        items = [ 'name:',
                  [ 'John',
                    'j is my nick'
                  ],
                  [ 'age:' ],
                  [ 26, ],
                  [ 'experience:' ],
                  [ '2000 THU',
                    '2006 sina',
                    '2010 other'
                  ],
        ]

        format_line(items, sep=' | ', aligns = 'llllll')

    outputs:

        name: | John         | age: | 26 | experience: | 2000 THU
              | j is my nick |      |    |             | 2006 sina
              |              |      |    |             | 2010 other

    '''

    aligns = [x for x in aligns] + [''] * len(items)
    aligns = aligns[:len(items)]
    aligns = ['r' if x == 'r' else x for x in aligns]

    items = [(x if type(x) in listtype else [x])
             for x in items]

    items = [[_to_str(y)
              for y in x]
             for x in items]

    maxHeight = max([len(x) for x in items] + [0])

    max_width = lambda x: max([y.__len__()
                               for y in x] + [0])

    widths = [max_width(x) for x in items]

    items = [(x + [''] * maxHeight)[:maxHeight]
             for x in items]

    lines = []
    for i in range(maxHeight):
        line = []
        for j in range(len(items)):
            width = widths[j]
            elt = items[j][i]

            actualWidth = elt.__len__()
            elt = utf8str(elt)

            if actualWidth < width:
                padding = ' ' * (width - actualWidth)
                if aligns[j] == 'l':
                    elt = elt + padding
                else:
                    elt = padding + elt

            line.append(elt)

        line = sep.join(line)

        lines.append(line)

    return "\n".join(lines)


def struct_repr(data, key=None):
    '''
    Render a data to a multi-line structural(yaml-like) representation.

        a = {
                1: 3,
                'x': {1:4, 2:5},
                'l': [1, 2, 3],
        }
        for l in struct_repr(a):
            print l

    Output:

        1 : 3
        l : - 1
            - 2
            - 3
        x : 1 : 4
            2 : 5
    '''

    if type(data) in listtype:

        if len(data) == 0:
            return ['[]']

        max_width = 0
        elt_lines = []
        for elt in data:
            sublines = struct_repr(elt)
            sublines_max_width = max([len(x) for x in sublines])

            if max_width < sublines_max_width:
                max_width = sublines_max_width

            elt_lines.append(sublines)

        lines = []
        for sublines in elt_lines:

            # - subline[0]
            #   subline[1]
            #   ...

            lines.append('- ' + sublines[0].ljust(max_width))

            for l in sublines[1:]:
                lines.append('  ' + l.ljust(max_width))

        return lines

    elif type(data) == types.DictType:

        if len(data) == 0:
            return ['{}']

        max_k_width = 0
        max_v_width = 0

        kvs = []

        for k, v in data.items():
            k = utf8str(k)
            sublines = struct_repr(v)
            sublines_max_width = max([len(x) for x in sublines])

            if max_k_width < len(k):
                max_k_width = len(k)

            if max_v_width < sublines_max_width:
                max_v_width = sublines_max_width

            kvs.append((k, sublines))

        kvs.sort(key=key)

        lines = []
        for k, sublines in kvs:

            # foo : sub-0
            #       sub-1
            #   b : sub-0
            #       sub-0

            lines.append(k.rjust(max_k_width) + ' : ' +
                         sublines[0].ljust(max_v_width))

            for l in sublines[1:]:
                lines.append(' '.rjust(max_k_width) +
                             '   ' + l.ljust(max_v_width))

        return lines

    else:
        return [utf8str(data)]


def _get_key_and_headers(keys, rows):

    if keys is None:

        if len(rows) == 0:
            keys = []
        else:
            r0 = rows[0]

            if type(r0) == types.DictType:
                keys = r0.keys()
                keys.sort()
            elif type(r0) in listtype:
                keys = [i for i in range(len(r0))]
            else:
                keys = ['']

    _keys = []
    column_headers = []

    for k in keys:

        if type(k) not in listtype:
            k = [k, k]

        _keys.append(k[0])
        column_headers.append(str(k[1]))

    return _keys, column_headers


def _get_colors(colors, col_n):

    if colors is None:
        colors = []

    colors = colors or ([None] * col_n)

    while len(colors) < col_n:
        colors.extend(colors)

    colors = colors[:col_n]

    return colors


def format_table(rows,
                 keys=None,
                 colors=None,
                 sep=' | ',
                 row_sep=None):

    keys, column_headers = _get_key_and_headers(keys, rows)
    colors = _get_colors(colors, len(keys))

    # element of lns is a mulit-column line
    # lns = [
    #         # line 1
    #         [
    #                 # column 1 of line 1
    #                 ['name:', # row 1 of column 1 of line 1
    #                  'foo',   # row 2 of column 1 of line 1
    #                 ],
    #
    #                 # column 2 of line 1
    #                 ['school:',
    #                  'foo',
    #                  'bar',
    #                 ],
    #         ],
    # ]

    # headers
    lns = [
        [[a + ': ']
         for a in column_headers]
    ]

    for row in rows:

        if row_sep is not None:
            lns.append([[None] for k in keys])

        if type(row) == types.DictType:

            ln = [struct_repr(row.get(k, ''))
                  for k in keys]

        elif type(row) in listtype:

            ln = [struct_repr(row[int(k)])
                  if len(row) > int(k) else ''
                  for k in keys]

        else:
            ln = [struct_repr(row)]

        lns.append(ln)

    get_max_width = lambda cols: max([len(utf8str(c[0]))
                                      for c in cols] + [0])

    max_widths = [get_max_width(cols) for cols in zip(*lns)]

    rows = []
    for row in lns:

        ln = []

        for i in range(len(max_widths)):
            color = colors[i]
            w = max_widths[i]

            ln.append([ColoredString(x.ljust(w), color)
                       if x is not None else row_sep * w
                       for x in row[i]])

        rows.append(format_line(ln, sep=sep))

    return rows


def _to_str(y):
    if isinstance(y, ColoredString):
        pass
    elif type(y) in (type(0), type(0L)):
        y = str(y)
    elif type(y) in (type([]), type(()), type({})):
        y = str(y)

    return y


def utf8str(s):

    if type(s) == type(u''):
        return s.encode('utf8')
    else:
        return str(s)


def common_prefix(a, *others, **options):

    recursive = options.get('recursive', True)
    for b in others:
        if type(a) != type(b):
            raise TypeError('a and b has different type: ' + repr((a, b)))
        a = _common_prefix(a, b, recursive)

    return a


def _common_prefix(a, b, recursive=True):

    rst = []
    for i, elt in enumerate(a):
        if i == len(b):
            break

        if type(elt) != type(b[i]):
            raise TypeError('a and b has different type: ' + repr((elt, b[i])))

        if elt == b[i]:
            rst.append(elt)
        else:
            break

    # Find common prefix of the last different element.
    #
    # string does not support nesting level reduction. It infinitely recurses
    # down.
    # And non-iterable element is skipped, such as int.
    i = len(rst)
    if recursive and i < len(a) and i < len(b) and not isinstance(a, basestring) and hasattr(a[i], '__len__'):

        last_prefix = _common_prefix(a[i], b[i])

        # discard empty tuple, list or string
        if len(last_prefix) > 0:
            rst.append(last_prefix)

    if isinstance(a, tuple):
        return tuple(rst)
    elif isinstance(a, list):
        return rst
    else:
        return ''.join(rst)


def colorize(percent, total=100, ptn='{0}'):
    if total > 0:
        color = fading_color(percent, total)
    else:
        color = fading_color(-total - percent, -total)
    return ColoredString(ptn.format(percent), color)


class ColoredString(object):

    def __init__(self, v, color=None, prompt=True):
        if type(color) in types.StringTypes:
            color = _named_colors[color]

        if isinstance(v, ColoredString):
            vs = ''.join([x[0] for x in v.elts])
            self.elts = [(vs, color)]
        else:
            self.elts = [(str(v), color)]

        self._prompt = prompt

    def __str__(self):
        rst = []
        for e in self.elts:
            if len(e[0]) == 0:
                continue

            if e[1] is None:
                val = e[0]
            else:
                _clr = '\033[38;5;' + str(e[1]) + 'm'
                _rst = '\033[0m'

                if self._prompt:
                    _clr = '\001' + _clr + '\002'
                    _rst = '\001' + _rst + '\002'

                val = _clr + str(e[0]) + _rst

            rst.append(val)

        return ''.join(rst)

    def __len__(self):
        return sum([len(x[0])
                    for x in self.elts])

    def __add__(self, other):
        prompt = self._prompt
        if isinstance(other, ColoredString):
            prompt = prompt or other._prompt

        c = ColoredString('', prompt=prompt)
        if isinstance(other, ColoredString):
            c.elts = self.elts + other.elts
        else:
            c.elts = self.elts[:] + [(str(other), None)]
        return c

    def __mul__(self, num):
        c = ColoredString('', prompt=self._prompt)
        c.elts = self.elts * num
        return c

    def __eq__(self, other):
        if not isinstance(other, ColoredString):
            return False
        return str(self) == str(other) and self._prompt == other._prompt

    def _find_sep(self, line, sep):
        ma = re.search(sep, line)
        if ma is None:
            return -1, 0

        return ma.span()

    def _recover_colored_str(self, colored_chars):
        rst = ColoredString('')
        n = len(colored_chars)
        if n == 0:
            return rst

        head = list(colored_chars[0])
        for ch in colored_chars[1:]:
            if head[1] == ch[1]:
                head[0] += ch[0]
            else:
                rst += ColoredString(head[0], head[1])
                head = list(ch)
        rst += ColoredString(head[0], head[1])

        return rst

    def _split(self, line, colored_chars, sep, maxsplit, keep_sep, keep_empty):
        rst = []
        n = len(line)
        i = 0
        while i < n:
            if maxsplit == 0:
                break

            s, e = self._find_sep(line[i:], sep)

            if s < 0:
                break

            edge = s
            if keep_sep:
                edge = e

            rst.append(self._recover_colored_str(colored_chars[i:i+edge]))

            maxsplit -= 1
            i += e

        if i < n:
            rst.append(self._recover_colored_str(colored_chars[i:]))

        # sep in the end
        # 'a b '  ->  ['a', 'b', '']
        elif keep_empty:
            rst.append(ColoredString(''))

        return rst

    def _separate_str_and_colors(self):
        colored_char = []
        line = ''
        for elt in self.elts:
            for c in elt[0]:
                colored_char.append((c, elt[1]))
            line += elt[0]

        return line, colored_char

    def splitlines(self, *args):
        # to verify arguments
        ''.splitlines(*args)

        sep = '\r(\n)?|\n'
        maxsplit = -1
        keep_empty = False
        keep_sep = False
        if len(args) > 0:
            keep_sep = args[0]

        line, colored_chars = self._separate_str_and_colors()

        return self._split(line, colored_chars, sep, maxsplit, keep_sep, keep_empty)

    def split(self, *args):
        # to verify arguments
        ''.split(*args)

        sep, maxsplit = (list(args) + [None, None])[:2]
        if maxsplit is None:
            maxsplit = -1
        keep_empty = True
        keep_sep = False

        line, colored_chars = self._separate_str_and_colors()

        i = 0
        if sep is None:
            sep = '\s+'
            keep_empty = False

            # to skip whitespaces at the beginning
            # ' a b'.split() -> ['a', 'b']
            n = len(line)
            while i < n and line[i] in string.whitespace:
                i += 1

        return self._split(line[i:], colored_chars[i:], sep, maxsplit, keep_sep, keep_empty)

    def join(self, iterable):
        rst = ColoredString('')
        for i in iterable:
            if len(rst) == 0:
                rst += i
            else:
                rst += self + i
        return rst


def fading_color(v, total):
    return _clrs[_fading_idx(v, total)]


def _fading_idx(v, total=100):
    l = len(_clrs)
    pos = int(v * l / (total + 0.0001) + 0.5)
    pos = min(pos, l - 1)
    pos = max(pos, 0)
    return pos


_clrs = [63, 67, 37, 36, 41, 46, 82, 118,
         154, 190, 226, 220, 214, 208, 202, 196]

_named_colors = {
    # by emergence levels
    'danger': _clrs[_fading_idx(100)],
    'warn': 3,
    'loaded': _clrs[_fading_idx(30)],
    'normal': 7,
    'optimal': _clrs[_fading_idx(0)],

    'dark': _clrs[1],

    # for human
    'blue': 67,
    'cyan': 37,
    'green': 46,
    'yellow': 226,
    'red': 196,
    'purple': 128,
    'white': 255,
}


def _make_colored_function(name):
    def _colored(v):
        return ColoredString(v, name)

    _colored.__name__ = name

    return _colored

for _func_name in _named_colors:
    setattr(sys.modules[__name__],
            _func_name, _make_colored_function(_func_name))


def break_line(linestr, width):
    lines = linestr.splitlines()
    rst = []

    space = ' '
    if isinstance(linestr, ColoredString):
        space = ColoredString(' ')

    for line in lines:
        words = line.split(' ')

        buf = words[0]
        for word in words[1:]:
            if len(word) + len(buf) + 1 > width:
                rst.append(buf)
                buf = word
            else:
                buf += space + word

        if buf != '':
            rst.append(buf)

    return rst



#!/usr/bin/env python2
# coding: utf-8

import thread
import time
import unittest

from pykit import cacheable


class TestLRU(unittest.TestCase):

    def _assert_lru_list(self, lru):

        size = lru.size
        item_head = lru.head
        item_tail = lru.tail
        for i in range(size):
            item_head = item_head['next']
            item_tail = item_tail['pre']

        self.assertIs(item_head, lru.tail)
        self.assertIs(item_tail, lru.head)

    def _assert_lru_items_order(self, lru, order_keys):

        item = lru.head['next']
        lru_keys = []
        while item is not None:
            lru_keys.append(item['key'])
            item = item['next']

        self.assertEqual(order_keys, lru_keys)

    def test_lru_timeout(self):

        cases = (
            ('k1', 'v1', 3, True),
            ('k2', 'v2', 1, False),
        )

        lru = cacheable.LRU(10, 2)
        for key, val, sleep_time, is_timeout in cases:
            lru[key] = val
            time.sleep(sleep_time)

            try:
                lru[key]
                self.assertFalse(is_timeout)

            except KeyError:
                self._assert_lru_list(lru)
                self.assertTrue(is_timeout)

    def test_lru_getitem(self):

        capacity = 4
        cases = (
            (3,
             [0, 1, 2],
             []),

            (4,
             [0, 1, 2, 3],
             []),

            (5,
             [0, 1, 2, 3, 4],
             []),

            (6,
             [0, 1, 2, 3, 4, 5],
             []),

            (7,
             [3, 4, 5, 6],
             [0, 1, 2]),
        )

        for insert_count, exist_items, cleanup_items in cases:
            lru = cacheable.LRU(capacity, 10)
            for i in range(insert_count):
                lru[i] = 'val%d' % (i)

            for i in range(insert_count):
                try:
                    val = lru[i]
                    self.assertEqual(val, 'val%d' % (i))
                    self.assertEqual(lru.tail['key'], i)
                    self.assertIn(i, exist_items)

                except KeyError:
                    self.assertIn(i, cleanup_items)

    def test_lru_setitem(self):

        capacity = 4
        cases = (
            (3,
             [0, 1, 2]),

            (4,
             [0, 1, 2, 3]),

            (5,
             [0, 1, 2, 3, 4]),

            (6,
             [0, 1, 2, 3, 4, 5]),

            # size of lru > capacity*1.5
            # clean items from head, until size=capacity
            (7,
             [3, 4, 5, 6]),
        )

        for insert_count, expect_order_keys in cases:
            lru = cacheable.LRU(capacity, 10)
            for i in range(insert_count):
                lru[i] = 'val'
                self._assert_lru_list(lru)

            self._assert_lru_items_order(lru, expect_order_keys)

    def test_lru_capacity(self):

        cases = (
                (0,
                 ((0, 0), (1, 0))),

                (1,
                 ((1, 1), (2, 1))),

                (10,
                 ((9, 9), (10, 10), (13, 13), (15, 15), (16, 10))),
        )

        for capacity, case in cases:
            for insert_count, expect_size in case:
                lru = cacheable.LRU(capacity, 60)
                for i in range(insert_count):
                    lru[i] = 'val'

                self.assertEqual(lru.size, expect_size)


class TestCacheable(unittest.TestCase):

    @cacheable.cache('method_cache_data', capacity=10, timeout=4, is_deepcopy=False)
    def _method_cache_data(self, key):
        data = need_cache_data.get(key, {})
        data['tm'] = time.time()
        return data

    def test_get_items_from_cache(self):

        cache_items = {}
        for key in need_cache_data.iterkeys():
            cache_items[key] = get_cache_data(key)

        time.sleep(0.1)
        for key in need_cache_data.iterkeys():
            self.assertEqual(get_cache_data(key), cache_items[key])

    def test_get_items_from_cache_use_method(self):

        cache_items = {}
        for key in need_cache_data.iterkeys():
            cache_items[key] = self._method_cache_data(key)

        time.sleep(0.1)
        for key in need_cache_data.iterkeys():
            self.assertEqual(self._method_cache_data(key), cache_items[key])

    def test_cache_item_timeout_and_cache_again(self):

        cases = (
            (2, False),
            (3, True),
        )

        tm = get_cache_data('key')['tm']
        for sleep_time, is_timeout in cases:
            time.sleep(sleep_time)
            if is_timeout:
                self.assertNotEqual(tm, get_cache_data('key')['tm'])
            else:
                self.assertEqual(tm, get_cache_data('key')['tm'])

    def test_get_deepcopy_item_from_cache(self):

        self.assertIsNot(get_deepcopy_of_cache_data('key1'),
                         get_deepcopy_of_cache_data('key1'))

    def test_get_concurrent_update_cache_data(self):

        result = {}
        def _store_result(key):
            result[key] = get_concurrent_update_cache_data()

        thread.start_new_thread(_store_result, ('key1',))
        time.sleep(1)
        thread.start_new_thread(_store_result, ('key2',))

        while True:
            if 'key2' in result:
                break
            time.sleep(0.1)

        self.assertIsNot(result['key1'], result['key2'])

    def test_get_mutext_update_cache_data(self):

        result = {}
        def _store_result(key):
            result[key] = get_mutex_update_cache_data()

        thread.start_new_thread(_store_result, ('key1',))
        time.sleep(1)
        thread.start_new_thread(_store_result, ('key2',))

        while True:
            if 'key2' in result:
                break
            time.sleep(0.1)

        self.assertIs(result['key1'], result['key2'])

    def test_generate_lru_key(self):

        cases = (
            ((),
             {},
             "[(), []]"),

            ((1),
             {1: 'val_1'},
             "[1, [(1, 'val_1')]]"),

            ((1),
             {'a': 'val_a'},
             "[1, [('a', 'val_a')]]"),

            (('c'),
             {'b': 'val_b'},
             "['c', [('b', 'val_b')]]"),

            ((1, 'c'),
             {'a': 'val_a', 'b': 'val_b'},
             "[(1, 'c'), [('a', 'val_a'), ('b', 'val_b')]]"),

            (('c', 1),
             {'b': 'val_b', 'a': 'val_a'},
             "[('c', 1), [('a', 'val_a'), ('b', 'val_b')]]"),
        )

        for args, argkv, expect_str in cases:
            self.assertEqual(cacheable.Cacheable()._arg_str(args, argkv),
                             expect_str)


@cacheable.cache('deepcopy_of_cache_data', capacity=100, timeout=60, is_deepcopy=True)
def get_deepcopy_of_cache_data(key):

    return need_cache_data.get(key, {})


@cacheable.cache('cache_data', capacity=100, timeout=4, is_deepcopy=False)
def get_cache_data(key):

    cache_data = need_cache_data.get(key, {})
    cache_data['tm'] = time.time()
    return cache_data

@cacheable.cache('concurrent_update_cache_data', capacity=100, timeout=60,
                    is_deepcopy=False, mutex_update=False)
def get_concurrent_update_cache_data():
    time.sleep(3)
    return {}

@cacheable.cache('mutex_update_cache_data', capacity=100, timeout=60,
                    is_deepcopy=False, mutex_update=True)
def get_mutex_update_cache_data():
    time.sleep(3)
    return {}

need_cache_data = {
    'key1': {'tm': 0},
    'key2': {'tm': 0},
    'key3': {'tm': 0},
    'key4': {'tm': 0},
}



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import os
import sys
import unittest

from pykit import modutil
from pykit import ututil

dd = ututil.dd

class TestModutil(unittest.TestCase):

    def setUp(self):

        sys.path.append(os.path.dirname(__file__))
        module_tree = [
                'root0',
                'root0.mod0',
                'root0.mod0.mod00',
                'root0.mod0.mod01',
                'root0.mod1',
                'root0.mod1.mod10',
                'root0.mod2',
                'root1',
                'root2',
                ]

        for module in module_tree:
            __import__(module)

    def tearDown(self):
        sys.path.remove(os.path.dirname(__file__))

    def test_submodules(self):

        test_cases = [
                (sys.modules['root0'], {
                                        'mod0': sys.modules['root0.mod0'],
                                        'mod1': sys.modules['root0.mod1'],
                                        'mod2': sys.modules['root0.mod2'], }, None),
                (sys.modules['root1'], {},                                    None),
                (sys.modules['root2'], None,                                  None),
                ({},                   None,                                  AttributeError),
            ]

        for root, rst_expected, error in test_cases:

            dd('case: ', root, rst_expected, error)

            try:
                rst = modutil.submodules(root)
            except error as e:
                self.assertEqual(type(e), error)
            else:
                dd('rst: ', rst)
                self.assertEqual(rst, rst_expected)

    def test_submodule_tree(self):

        test_cases = [
                ( sys.modules['root0'],
                  {
                    'mod0': {'module': sys.modules['root0.mod0'],
                            'children': {
                                'mod00': {'module': sys.modules['root0.mod0.mod00'],
                                          'children': {},
                                },
                                'mod01': {'module': sys.modules['root0.mod0.mod01'],
                                          'children': None,
                                },
                            },
                           },
                    'mod1': {'module': sys.modules['root0.mod1'],
                            'children': {
                                'mod10': {'module': sys.modules['root0.mod1.mod10'],
                                          'children': None,
                                },
                            },
                           },
                   'mod2': {'module': sys.modules['root0.mod2'],
                            'children': None,
                           },
                  },
                  None, ),
                ( sys.modules['root1'],
                  {},
                  None, ),
                ( sys.modules['root2'],
                  None,
                  None, ),
                ( {},
                  None,
                  AttributeError,),
                ]

        for root, rst_expected, error in test_cases:

            dd('case: ', root, rst_expected, error)

            try:
                rst = modutil.submodule_tree(root)
            except error as e:
                self.assertEqual(type(e), error)
            else:
                dd('rst: ', rst)
                self.assertEqual(rst, rst_expected)

    def test_submodule_leaf_tree(self):

        test_cases = [
                ( sys.modules['root0'],
                  {
                      'mod0': { 'mod00': {},
                                'mod01':sys.modules['root0.mod0.mod01'],
                              },
                      'mod1': { 'mod10': sys.modules['root0.mod1.mod10'], },
                      'mod2': sys.modules['root0.mod2'],
                  },
                  None, ),
                ( sys.modules['root1'],
                  {},
                  None, ),
                ( sys.modules['root2'],
                  None,
                  None, ),
                ( {},
                  None,
                  AttributeError, ),
                ]

        for root, rst_expected, error in test_cases:

            dd('case: ', root, rst_expected, error)

            try:
                rst = modutil.submodule_leaf_tree(root)
            except error as e:
                self.assertEqual(type(e), error)
            else:
                dd('rst: ', rst)
                self.assertEqual(rst, rst_expected)



#!/usr/bin/env python2
# coding: utf-8

import re
import socket
import uuid
from collections import defaultdict
from collections import namedtuple

import psutil

from pykit import fsutil
from pykit import net
from pykit import strutil

from .idbase import IDBase


class DriveIDError(Exception):
    pass


class ServerID(str):

    @classmethod
    def validate(cls, server_id):
        if not isinstance(server_id, basestring):
            return False

        return re.match("^[0-9a-f]{12}$", server_id) is not None

    @classmethod
    def local_server_id(self):
        return ServerID('%012x' % uuid.getnode())


class DriveID(namedtuple('_DriveID', 'server_id mountpoint_index'), IDBase):

    _tostr_fmt = '{server_id}0{mountpoint_index:0>3}'

    @classmethod
    def validate(cls, drive_id):
        if not isinstance(drive_id, basestring):
            return False

        if len(drive_id) != 16:
            return False

        server_id = drive_id[:12]
        padding = drive_id[12:13]
        mp_idx = drive_id[13:]

        return (ServerID.validate(server_id)
                and padding == '0'
                and re.match("^[0-9]{3}$", mp_idx) is not None)

    @classmethod
    def parse(cls, drive_id):
        if not DriveID.validate(drive_id):
            raise DriveIDError('invalid drive id: {d}'.format(d=drive_id))

        return DriveID(ServerID(drive_id[:12]),
                       int(drive_id[13:]))


def _make_mountpoints_info():
    mps = fsutil.get_all_mountpoint()
    prt_by_mp = fsutil.get_disk_partitions()

    mps_info = defaultdict(dict)
    for m in mps:
        p = prt_by_mp.get(m, None)
        if p is None:
            continue

        capacity = fsutil.get_path_usage(m)['total']

        mps_info[m]['fs'] = p['fstype']
        mps_info[m]['capacity'] = capacity

    return dict(mps_info)


def _get_allocated_drive(allocated_drive_pre, mountpoints):
    rst = {}
    for m in mountpoints:
        if not m.startswith(allocated_drive_pre):
            continue

        rst[m] = {'status': 'normal'}

    return rst


def make_serverrec(idc, idc_type, roles, allocated_drive_pre, **argkv):
    serverrec = {}

    ips = net.get_host_ip4(exclude_prefix="docker")
    inn_ips = net.choose_inn(ips)
    pub_ips = net.choose_pub(ips)

    memory = psutil.virtual_memory().total
    cpu_info = {}
    # count of logical cpus
    cpu_info['count'] = psutil.cpu_count()
    # Mhz
    if hasattr(psutil, 'cpu_freq'):
        cpu_info['frequency'] = psutil.cpu_freq().max

    serverrec['server_id'] = str(ServerID())
    serverrec['pub_ips'] = pub_ips
    serverrec['inn_ips'] = inn_ips
    serverrec['hostname'] = socket.gethostname()
    serverrec['memory'] = memory
    serverrec['cpu'] = cpu_info
    serverrec['idc'] = idc
    serverrec['idc_type'] = idc_type
    serverrec['roles'] = roles

    mps = _make_mountpoints_info()
    serverrec['mountpoints'] = mps
    serverrec['next_mount_index'] = 1
    serverrec['allocated_drive'] = _get_allocated_drive(allocated_drive_pre, mps)
    serverrec.update(argkv)

    return serverrec


def get_serverrec_str(serverrec):
    rst = []

    for k in ('server_id', 'idc', 'idc_type', 'roles'):
        rst.append('{k}: {v}'.format(k=k, v=serverrec[k]))

    rst.append('mountpoints_count: {cnt}'.format(
               cnt=len(serverrec['mountpoints'])))

    rst.append('allocated_drive_count: {cnt}'.format(
        cnt=len(serverrec['allocated_drive'])))

    return '; '.join(rst)


def validate_idc(idc):
    if not isinstance(idc, basestring):
        return False

    if len(idc) > 0 and not idc.startswith('.'):
        return False

    names = idc.split('.')
    for n in names:
        if re.match("^[0-9a-zA-Z]*$", n) is None:
            return False

    return True


def idc_distance(idc_a, idc_b):
    if idc_a == idc_b:
        return 0

    pre_cnt = len(strutil.common_prefix(idc_a.split('.'), idc_b.split('.')))
    return 32 / (2 ** pre_cnt)



import time

import daemonize

fn = '/tmp/foo'
pidfn = '/tmp/test_daemonize.pid'


def write_file(fn, cont):
    try:
        with open(fn, 'w') as f:
            f.write(cont)
    except Exception as e:
        print repr(e) + ' while write_file:' + fn
        raise


def run():

    write_file(fn, 'foo-before')

    time.sleep(1)

    write_file(fn, 'foo-after')


daemonize.daemonize_cli(run, pidfn)



import os
import sys
import textwrap
import unittest
import subprocess32
from test import test_support
from test.script_helper import assert_python_ok

class TestTool(unittest.TestCase):
    data = """

        [["blorpie"],[ "whoops" ] , [
                                 ],\t"d-shtaeou",\r"d-nthiouh",
        "i-vhbjkhnth", {"nifty":87}, {"morefield" :\tfalse,"field"
            :"yes"}  ]
           """

    expect = textwrap.dedent("""\
    [
        [
            "blorpie"
        ],
        [
            "whoops"
        ],
        [],
        "d-shtaeou",
        "d-nthiouh",
        "i-vhbjkhnth",
        {
            "nifty": 87
        },
        {
            "field": "yes",
            "morefield": false
        }
    ]
    """)

    def test_stdin_stdout(self):
        proc = subprocess32.Popen(
                (sys.executable, '-m', 'json.tool'),
                stdin=subprocess32.PIPE, stdout=subprocess32.PIPE)
        out, err = proc.communicate(self.data.encode())
        self.assertEqual(out.splitlines(), self.expect.encode().splitlines())
        self.assertEqual(err, None)

    def _create_infile(self):
        infile = test_support.TESTFN
        with open(infile, "w") as fp:
            self.addCleanup(os.remove, infile)
            fp.write(self.data)
        return infile

    def test_infile_stdout(self):
        infile = self._create_infile()
        rc, out, err = assert_python_ok('-m', 'json.tool', infile)
        self.assertEqual(out.splitlines(), self.expect.encode().splitlines())
        self.assertEqual(err, b'')

    def test_infile_outfile(self):
        infile = self._create_infile()
        outfile = test_support.TESTFN + '.out'
        rc, out, err = assert_python_ok('-m', 'json.tool', infile, outfile)
        self.addCleanup(os.remove, outfile)
        with open(outfile, "r") as fp:
            self.assertEqual(fp.read(), self.expect)
        self.assertEqual(out, b'')
        self.assertEqual(err, b'')



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as fh:
    install_requires=[line.strip('\n') for line in fh.readlines() if line.strip() and not line.strip().startswith('#')]

setuptools.setup(
    name="pykit-bsc",
    version="0.0.2",
    author="Zhang Yanpo",
    author_email="drdr.xp@gmail.com",
    description="A collection of toolkit lib for distributed system development in python",
    license="MIT",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/bsc-s2/pykit",
    packages=setuptools.find_packages(),
    install_requires=install_requires,
    python_requires='>=2.7',
    classifiers=(
        "Programming Language :: Python :: 2",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
)



import logging
import time

logger = logging.getLogger(__name__)


def enqueue_log_entry(ts_cache, queue):
    # ts_cache = {
    #     'source_file_xxx': {
    #         123: {  # 123 is line number
    #             ...
    #         },
    #         ...
    #     },
    #     ...
    # }
    for _, source_file_cache in ts_cache.iteritems():
        for _, log_entry in source_file_cache.iteritems():
            queue.put(log_entry)


def flush_cache(log_cache, queue):
    ts_list = log_cache.keys()

    if len(ts_list) == 0:
        return

    ts_list.sort()

    # keep the latest log if it is generated in about 2 second.
    if time.time() - ts_list[-1] < 2:
        ts_list = ts_list[:-1]

    # only enqueue logs of latest 3 ts.
    for log_ts in ts_list[-3:]:
        enqueue_log_entry(log_cache[log_ts], queue)

    for log_ts in ts_list:
        del log_cache[log_ts]


def one_flush(context):

    for log_name in context['cache'].keys():
        log_cache = context['cache'][log_name]
        log_stat = context['stat'][log_name]

        try:
            flush_cache(log_cache, context['queue'])
            log_stat['flush_cache_error'] = None

        except Exception as e:
            logger.exception('failed to flush cache of: %s, %s' %
                             (log_name, repr(e)))
            log_stat['flush_cache_error'] = repr(e)


def run(context):
    while True:
        start_time = time.time()

        with context['cache_lock']:
            one_flush(context)

        time_used = time.time() - start_time

        logger.info('flush at: %f, time used: %f' %
                    (start_time, time_used))

        time.sleep(1)



import ctypes
import datetime
import logging
import threading
import time

import pytz
import tzlocal

from pykit import threadutil
from pykit import timeutil
from pykit import utfjson

SYS_gettid = 186

logger = logging.getLogger(__name__)


class NextFireTimeError(Exception):
    pass


class JobExistError(Exception):
    pass


SECOND = {
    'second': 1,
    'minute': 60,
    'hour': 60 * 60,
    'day': 60 * 60 * 24,
    'week': 60 * 60 * 24 * 7,
    'month': 60 * 60 * 24 * 31,
}


def get_time_info(ts=None, tz=None):
    if ts is None:
        ts = time.time()

    if tz is None:
        tz = tzlocal.get_localzone()

    dt = datetime.datetime.fromtimestamp(ts, tz)

    time_info = {
        'ts': ts,
        'string': str(dt),
    }

    return time_info


def get_next_fire_time(conf, last_fire_ts):
    n, unit = conf['every']
    interval = n * SECOND[unit]

    next_ts = last_fire_ts + interval

    at = conf.get('at')

    if at is None:
        return get_time_info(next_ts)

    timezone_name = conf.get('timezone')

    if timezone_name is None:
        tz = tzlocal.get_localzone()
    else:
        tz = pytz.timezone(timezone_name)

    next_dt = datetime.datetime.fromtimestamp(next_ts, tz)
    actual_next_dt = next_dt.replace(**at)

    actual_next_ts = timeutil.datetime_to_ts(actual_next_dt)

    next_fire_time = get_time_info(actual_next_ts)

    if actual_next_ts <= last_fire_ts:
        msg = ('next fire time: %s, is before last fire ts: %f, conf: %s' %
               (repr(next_fire_time), last_fire_ts, repr(conf)))
        raise NextFireTimeError(msg)

    return next_fire_time


class JobScheduler(object):

    def __init__(self, jobs, dump_status=None, reload_status=None,
                 should_job_run=None):
        self.lock = threading.RLock()
        self.jobs = jobs
        self.status = {}

        self.dump_status = dump_status
        self.reload_status = reload_status
        self.should_job_run = should_job_run

        if self.reload_status is not None:
            try:
                status = self.reload_status()
                logger.info('loaded job_status: %s', repr(status))

                if status is not None:
                    self.status = status

            except Exception as e:
                logger.exception('failed to reload job status: %s' % repr(e))

    def append_log(self, job_status, job_conf, log_msg):
        job_status['log'].append(log_msg)

        # log at start and end, add keep 3 more entries.
        n = job_conf['concurrence_n'] * 2 + 3
        job_status['log'] = job_status['log'][-n:]

    def run_job(self, curr_time, job_name, job_conf, job_status):
        thread_id = None
        thread_ident = threading.current_thread().ident

        try:
            libc = ctypes.cdll.LoadLibrary('libc.so.6')
            thread_id = libc.syscall(SYS_gettid)
        except Exception as e:
            logger.info('failed to get thread id: %s' + repr(e))

        with self.lock:
            job_status['active_threads'][thread_ident] = {
                'start_time': curr_time,
                'thread_ident': thread_ident,
                'thread_id': thread_id,
            }

            log_msg = 'thread: %s-%s started at: %s' % (
                thread_ident, thread_id, curr_time)
            self.append_log(job_status, job_conf, log_msg)

        try:
            job_conf['func'](*job_conf['args'], **job_conf['kwargs'])

            with self.lock:
                log_msg = 'thread: %s-%s finished at: %s' % (
                    thread_ident, thread_id, get_time_info(ts=time.time()))
                self.append_log(job_status, job_conf, log_msg)

                del(job_status['active_threads'][thread_ident])

        except Exception as e:
            logger.exception('failed to run job: %s, %s' %
                             (job_name, repr(e)))

            with self.lock:
                log_msg = 'thread: %s-%s failed at: %s' % (
                    thread_ident, thread_id, get_time_info(ts=time.time()))
                self.append_log(job_status, job_conf, log_msg)

                del(job_status['active_threads'][thread_ident])
                job_status['message'] = repr(e)

    def fire(self, curr_time, job_name, job_conf, job_status):
        thread_n = len(job_status['active_threads'])

        if thread_n >= job_conf['concurrence_n']:
            log_msg = 'at time: %s, already have %d threads for job: %s' % (
                curr_time, thread_n, job_name)
            self.append_log(job_status, job_conf, log_msg)

            logger.error('too many threads for job: %s' % job_name)
            return

        logger.info('at time: %s, start to run job: %s' %
                    (curr_time, job_name))

        threadutil.start_daemon_thread(
            self.run_job, args=(curr_time, job_name, job_conf, job_status))

        job_status['message'] = ''

    def schedule_one_job(self, curr_time, job_name, job_conf):
        if 'concurrence_n' not in job_conf:
            job_conf['concurrence_n'] = 1

        if job_name not in self.status:
            self.status[job_name] = {
                'active_threads': {},
                'log': [],
            }

        job_status = self.status[job_name]

        job_status['curr_time'] = curr_time

        if self.should_job_run is not None:
            should_run = self.should_job_run(job_conf)

            if not should_run:
                job_status['message'] = 'this job do not need to run'
                return

        if 'next_fire_time' not in job_status:
            if job_conf.get('at') is None:
                job_status['fire_time'] = curr_time

                self.fire(curr_time, job_name, job_conf, job_status)

                job_status['next_fire_time'] = get_next_fire_time(
                    job_conf, curr_time['ts'])

            else:
                n, unit = job_conf['every']
                interval = n * SECOND[unit]

                next_fire_time = get_next_fire_time(
                    job_conf, curr_time['ts'] - interval)

                if next_fire_time['ts'] < curr_time['ts']:
                    next_fire_time = get_next_fire_time(
                        job_conf, curr_time['ts'])

                job_status['next_fire_time'] = next_fire_time

        if curr_time['ts'] >= job_status['next_fire_time']['ts']:
            job_status['fire_time'] = curr_time

            self.fire(curr_time, job_name, job_conf, job_status)

            job_status['next_fire_time'] = get_next_fire_time(
                job_conf, job_status['next_fire_time']['ts'])

    def _schedule(self, curr_time):
        for job_name, job_conf in self.jobs.iteritems():
            try:
                self.schedule_one_job(curr_time, job_name, job_conf)

            except Exception as e:
                logger.exception('failed to schedule job %s: %s' %
                                 (job_name, repr(e)))

    def schedule(self):
        for job_name, job_status in self.status.items():
            if job_name not in self.jobs:
                del(self.status[job_name])
                continue

            if len(job_status['active_threads']) > 0:
                msg = 'threads aborted by restart: %s' % (
                    job_status['active_threads'])
                self.status[job_name]['message'] = msg
                self.status[job_name]['active_threads'] = {}

        while True:
            curr_time = get_time_info(time.time())

            with self.lock:
                self._schedule(curr_time)

            if self.dump_status is not None:
                try:
                    self.dump_status(self.status)
                except Exception as e:
                    logger.exception('failed to dump job status: %s' % repr(e))

            for job_name, job_status in self.status.items():
                logger.info('status of job %s, %s' %
                            (job_name, utfjson.dump(job_status)))

            end_time = time.time()

            logger.info('scheduled at: %s, time used: %f' %
                        (repr(curr_time), end_time - curr_time['ts']))

            to_sleep = 60 - (end_time % 60) + 1
            time.sleep(to_sleep)

    def add_job(self, job_name, job_conf):
        with self.lock:
            if job_name in self.jobs:
                raise JobExistError('job: %s already exists' % job_name)

            self.jobs[job_name] = job_conf

    def delete_job(self, job_name):
        with self.lock:
            if job_name in self.jobs:
                del(self.jobs[job_name])

            if job_name in self.status:
                del(self.status[job_name])



import hashlib
import logging
import os
import threading
import time
import types
import uuid
from collections import namedtuple

from kazoo import security
from kazoo.client import KazooClient
from kazoo.exceptions import NoNodeError

from pykit import config
from pykit import net
from pykit import utfjson

from .exceptions import ZKWaitTimeout

logger = logging.getLogger(__name__)

PERM_TO_LONG = {
    'c': 'create',
    'd': 'delete',
    'r': 'read',
    'w': 'write',
    'a': 'admin',
}

PERM_TO_SHORT = {
    'create': 'c',
    'delete': 'd',
    'read': 'r',
    'write': 'w',
    'admin': 'a',
}

# We assumes that ip does not change during process running.
# Display intra ip if presents, or display pub ip.
host_ip4 = net.ips_prefer(net.get_host_ip4(), net.INN)


class PermTypeError(Exception):
    pass


def lock_data(node_id=None):
    # deprecated
    return lock_id(node_id=node_id)


def lock_id(node_id=None):
    """
    Embed lock holder information into the zk node data for the lock.

    `node_id` is a user defined identifier of a host.
    """

    if node_id is None:
        node_id = config.zk_node_id

    ip = (host_ip4 + ['unknownip'])[0]

    seq = [node_id, ip, str(os.getpid()), str(uuid.uuid4()).replace('-', '')]

    return '-'.join(seq)


def parse_lock_data(data_str):
    # deprecated
    return parse_lock_id(data_str)


def parse_lock_id(data_str):
    """
    Parse string generated by lock_id()
    """

    node_id, ip, process_id, _uuid = (data_str.split('-', 3) + ([None] * 4))[:4]

    if type(process_id) in types.StringTypes and process_id.isdigit():
        process_id = int(process_id)
    else:
        process_id = None

    rst = {
        'node_id': node_id,
        'ip': ip,
        'process_id': process_id,
        'uuid': _uuid,
        'txid': None
    }

    if node_id.startswith('txid:'):
        rst['txid'] = node_id.split(':', 1)[1]

    return rst


def make_digest(acc):
    # acc = "username:password"

    digest = hashlib.sha1(acc).digest().encode('base64').strip()
    return digest


def make_acl_entry(username, password, permissions):

    perms = ''
    for c in permissions:
        if c not in PERM_TO_LONG:
            raise PermTypeError(c)
        perms += c

    return "digest:{username}:{digest}:{permissions}".format(
        username=username,
        digest=make_digest(username + ":" + password),
        permissions=perms)


def perm_to_long(short, lower=True):

    rst = []

    for c in short:
        c = c.lower()
        if c not in PERM_TO_LONG:
            raise PermTypeError(c)

        rst.append(PERM_TO_LONG[c])

    if not lower:
        rst = [x.upper() for x in rst]

    return rst


def perm_to_short(lst, lower=True):

    rst = ''

    for p in lst:
        p = p.lower()
        if p not in PERM_TO_SHORT:
            raise PermTypeError(p)

        rst += PERM_TO_SHORT[p]

    if not lower:
        rst = rst.upper()

    return rst


def make_kazoo_digest_acl(acl):

    # acl = (('xp', '123', 'cdrwa'),
    #        ('foo', 'passw', 'rw'),
    # )

    if acl is None:
        return None

    rst = []
    for name, passw, perms in acl:
        perm_dict = {p: True
                     for p in perm_to_long(perms)}
        acl_entry = security.make_digest_acl(name, passw, **perm_dict)
        rst.append(acl_entry)

    return rst


def parse_kazoo_acl(acls):

    # acls = [ACL(perms=31,
    #            acl_list=['ALL'],
    #            id=Id(scheme='digest', id=u'user:+Ir5sN1lGJEEs8xBZhZXK='))]

    rst = []
    for acl in acls:
        if 'ALL' in acl.acl_list:
            acl_list = 'cdrwa'
        else:
            acl_list = perm_to_short(acl.acl_list)

        rst.append((acl.id.scheme, acl.id.id.split(':')[0], acl_list))

    return rst


def is_backward_locking(locked_keys, key):

    locked_keys = sorted(locked_keys)
    assert key not in locked_keys, 'must not re-lock a key'

    if len(locked_keys) == 0:
        is_backward = False
    else:
        is_backward = key < locked_keys[-1]

    return is_backward


def _init_node(zkcli, parent_path, node, val, acl, users):

    path = parent_path + '/' + node

    if acl is None:
        acls = zkcli.get_acls(parent_path)[0]
    else:
        acls = [(user, users[user], perms) for user, perms in acl.items()]
        acls = make_kazoo_digest_acl(acls)

    if zkcli.exists(path) is None:
        zkcli.create(path, value=val, acl=acls)
    else:
        zkcli.set_acls(path, acls)

    return path


def init_hierarchy(hosts, hierarchy, users, auth):

    zkcli = KazooClient(hosts)
    zkcli.start()

    scheme, name, passw = auth
    zkcli.add_auth(scheme, name + ':' + passw)

    def _init_hierarchy(hierarchy, parent_path):

        if len(hierarchy) == 0:
            return

        for node, attr_children in hierarchy.items():
            val = attr_children.get('__val__', {})
            val = utfjson.dump(val)
            acl = attr_children.get('__acl__')

            path = _init_node(zkcli, parent_path, node, val, acl, users)
            children = {k: v
                        for k, v in attr_children.items()
                        if k not in ('__val__', '__acl__')
                        }

            _init_hierarchy(children, path)

    _init_hierarchy(hierarchy, '/')
    zkcli.stop()


NeedWait = namedtuple('NeedWait', [])


def _conditioned_get_loop(zkclient, path, conditioned_get, timeout=None, **kwargs):

    if timeout is None:
        timeout = 86400 * 365

    expire_at = time.time() + timeout
    lck = threading.RLock()
    maybe_available = threading.Event()
    maybe_available.clear()

    def set_available():
        with lck:
            maybe_available.set()

    def on_connection_change(state):
        # notify it to re-get, then raise Connection related error
        logger.info('connection state change: {0}'.format(state))
        set_available()

    zkclient.add_listener(on_connection_change)

    try:
        while True:

            with lck:
                it = conditioned_get(zkclient, path, **kwargs)
                it.next()
                rst = it.send((NeedWait, set_available))

                if rst is NeedWait:
                    maybe_available.clear()
                else:
                    return rst

            if maybe_available.wait(expire_at - time.time()):
                continue

            raise ZKWaitTimeout("timeout({timeout} sec)"
                                " waiting for {path} to satisfy: {cond}".format(
                                    timeout=timeout,
                                    path=path,
                                    cond=str(kwargs)))
    finally:
        try:
            zkclient.remove_listener(on_connection_change)
        except Exception as e:
            logger.info(repr(e) + ' while removing on_connection_change')


def make_conditioned_getter(conditioned_get):

    def _wrap(zkclient, path, timeout=None, **kwargs):
        return _conditioned_get_loop(zkclient, path, conditioned_get, timeout=timeout, **kwargs)

    _wrap.__name__ = conditioned_get.__name__

    return _wrap


@make_conditioned_getter
def get_next(zkclient, path, version=-1):

    NeedWait, set_available = yield

    val, zstat = zkclient.get(path, watch=lambda watchevent: set_available())
    if zstat.version > version:
        yield val, zstat

    yield NeedWait


@make_conditioned_getter
def wait_absent(zkclient, path):

    NeedWait, set_available = yield

    try:
        zkclient.get(path, watch=lambda watchevent: set_available())
    except NoNodeError as e:
        logger.info(repr(e) + ' found, return')
        yield None

    yield NeedWait




import logging

from pykit import rangeset

logger = logging.getLogger(__name__)


def load(gtid_set_str):

    # https://dev.mysql.com/doc/refman/5.6/en/replication-gtids-concepts.html
    #
    # gtid_set:
    #     uuid_set [, uuid_set] ...
    #     | ''
    # uuid_set:
    #     uuid:interval[:interval]...
    # uuid:
    #     hhhhhhhh-hhhh-hhhh-hhhh-hhhhhhhhhhhh
    # h:
    #     [0-9|A-F]
    # interval:
    #     n[-n]
    #     (n >= 1)
    #
    # <uuid>:<start>-<end>:<start-2>-<end2>,\n
    # <uuid>:<start>,\n
    # <uuid>:<start>-<end>,\n
    # ...
    #
    # 1fcc256b-84b0-11e6-acaf-00163e03bb03:1-326,\n
    # xxxxxxxx-84b0-11e6-acaf-00163e03bb03:1

    # {
    #   <uuid>: IntIncRangeSet([[1, 1], [10, 30]]),
    #   <uuid>: IntIncRangeSet([[1, 1], [10, 30]]),
    #   ...
    # }
    _set = gtid_set_str.split(',')

    logger.debug('gtid set: ' + repr(_set))

    gtid_set = {}

    for uuid_set in _set:

        uuid_set = uuid_set.strip()

        uuid_intervals = uuid_set.split(':')
        uuid = uuid_intervals[0]
        intervals = uuid_intervals[1:]

        gtid_set[uuid] = []

        ranges = []

        for interval in intervals:

            interval = interval.split('-')

            # single transaction range is just xxx:1, not xxx:1-2
            if len(interval) == 1:
                interval = [interval[0], interval[0]]

            interval = (int(interval[0]), int(interval[1]))

            ranges.append(interval)

        gtid_set[uuid] = rangeset.IntIncRangeSet(ranges)

    logger.debug('gtid_set parsed: ' + repr(gtid_set))

    return gtid_set


def dump(gtid_set, line_break='\n'):

    rst = []

    for uuid, ranges in gtid_set.items():
        uuid_set_str = [uuid, ]
        for rng in ranges:
            if rng.length() == 1:
                itv_str = '{0}'.format(rng[0])
            else:
                itv_str = '{0}-{1}'.format(*rng)

            uuid_set_str.append(itv_str)

        rst.append(':'.join(uuid_set_str))

    rst.sort()
    return (',' + line_break).join(rst)


def compare(a, b):

    rst = {
        'onlyleft': {
            'length': 0,
            'gtidset': {},
        },
        'onlyright': {
            'length': 0,
            'gtidset': {},
        },
    }

    for uuid, intervals in a.items():

        ll = rangeset.substract(intervals, b.get(uuid, rangeset.IntIncRangeSet([])))
        if ll.length() > 0:
            rst['onlyleft']['gtidset'][uuid] = ll
            rst['onlyleft']['length'] += ll.length()

    for uuid, intervals in b.items():

        ll = rangeset.substract(intervals, a.get(uuid, rangeset.IntIncRangeSet([])))
        if ll.length() > 0:
            rst['onlyright']['gtidset'][uuid] = ll
            rst['onlyright']['length'] += ll.length()

    return rst



import logging
import os

from pykit import fsutil
from pykit.cgrouparch import cgroup_util

logger = logging.getLogger(__name__)


def set_cgroup(cgroup_path, conf):
    if 'share' not in conf:
        share = '1024'
    else:
        share = str(conf['share'])

    share_file = os.path.join(cgroup_path, 'cpu.shares')

    fsutil.write_file(share_file, share, fsync=False)
    logger.info('write: %s to file: %s' % (share, share_file))

    pids = conf.get('pids')

    if pids is None:
        return

    cgroup_util.add_pids(cgroup_path, pids)
    logger.info('add pids: %s to cgroup: %s' % (repr(pids), cgroup_path))


def reset_statistics(cgroup_path):
    usage_file = os.path.join(cgroup_path, 'cpuacct.usage')
    fsutil.write_file(usage_file, '0', fsync=False)


def account(cgroup_path):
    usage_file = os.path.join(cgroup_path, 'cpuacct.usage')

    usage = int(fsutil.read_file(usage_file).strip())

    return {
        'usage': usage,
    }



import logging

from kazoo.exceptions import BadVersionError

from pykit import txutil

from .zkconf import kazoo_client_ext

logger = logging.getLogger(__name__)


def cas_loop(zkclient, path, json=True):

    zkclient, owning_zk = kazoo_client_ext(zkclient, json=json)

    def setter(path, val, zstat):
        zkclient.set(path, val, version=zstat.version)

    try:
        for curr in txutil.cas_loop(zkclient.get, setter, args=(path, ),
                                    conflicterror=BadVersionError):
            yield curr
    finally:
        if owning_zk:
            try:
                zkclient.stop()
            except Exception as e:
                logger.info(repr(e) + ' while stop owning zkclient')



#!/usr/bin/env python2
# coding: utf-8

import httplib
import logging
import socket
import time
import urllib
import urlparse

from pykit import http
from pykit import utfjson

logger = logging.getLogger(__name__)


class EtcdException(Exception):
    pass


class EtcdInternalError(EtcdException):
    pass


class NoMoreMachineError(EtcdException):
    pass


class EtcdReadTimeoutError(EtcdException):
    pass


class EtcdRequestError(EtcdException):
    pass


class EtcdResponseError(EtcdException):
    pass


class EtcdIncompleteRead(EtcdResponseError):
    pass


class EtcdSSLError(EtcdException):
    pass


class EtcdWatchError(EtcdException):
    pass


class EtcdKeyError(EtcdException, KeyError):
    pass


class EtcdValueError(EtcdException, ValueError):
    pass


class EcodeKeyNotFound(EtcdKeyError):
    pass


class EcodeTestFailed(EtcdValueError):
    pass


class EcodeNotFile(EtcdKeyError):
    pass


class EcodeNotDir(EtcdKeyError):
    pass


class EcodeNodeExist(EtcdKeyError):
    pass


class EcodeRootROnly(EtcdValueError):
    pass


class EcodeDirNotEmpty(EtcdValueError):
    pass


class EcodePrevValueRequired(EtcdValueError):
    pass


class EcodeTTLNaN(EtcdValueError):
    pass


class EcodeIndexNaN(EtcdValueError):
    pass


class EcodeInvalidField(EtcdValueError):
    pass


class EcodeInvalidForm(EtcdValueError):
    pass


class EcodeInscientPermissions(EtcdException):
    pass


def list_type(x):
    if isinstance(x, (list, tuple)):
        return True

    return False


class EtcdError(object):
    error_exceptions = {
        100: EcodeKeyNotFound,
        101: EcodeTestFailed,
        102: EcodeNotFile,
        103: EtcdException,
        104: EcodeNotDir,
        105: EcodeNodeExist,
        106: EtcdKeyError,
        107: EcodeRootROnly,
        108: EcodeDirNotEmpty,
        110: EcodeInscientPermissions,
        200: EtcdValueError,
        201: EcodePrevValueRequired,
        202: EcodeTTLNaN,
        203: EcodeIndexNaN,
        209: EcodeInvalidField,
        210: EcodeInvalidForm,
        300: EtcdInternalError,
        301: EtcdInternalError,
        400: EtcdWatchError,
        401: EtcdWatchError,
        500: EtcdInternalError,
    }

    @classmethod
    def handle(cls, response):

        body = response.data

        e = {}
        e['status'] = response.status
        e['headers'] = response.headers
        e['response'] = body

        try:
            r = utfjson.load(body)
        except ValueError:
            r = {"message": "response body is not json", "cause": str(body)}

        ecode = r.get('errorCode')
        default_exc = EtcdException
        if response.status == 404:
            ecode = 100
        elif response.status == 401:
            ecode = 110
        elif response.status >= 500:
            default_exc = EtcdResponseError

        exc = cls.error_exceptions.get(ecode, default_exc)
        if ecode in cls.error_exceptions:
            msg = "{msg} : {cause}".format(msg=r.get('message'),
                                           cause=r.get('cause'))
        else:
            msg = "Unable to decode server response"

        e['message'] = msg

        raise exc(e)


class EtcdKeysResult(object):

    _node_props = {
        'key': None,
        'value': None,
        'expiration': None,
        'ttl': None,
        'modifiedIndex': None,
        'createdIndex': None,
        'newKey': False,
        'dir': False,
    }

    def __init__(self, action=None, node=None, prevNode=None, **argkv):

        self.action = action
        for key, default in self._node_props.items():
            if node is not None and key in node:
                setattr(self, key, node[key])
            else:
                setattr(self, key, default)

        self._children = []
        if self.dir and 'nodes' in node:
            self._children = node['nodes']

        if prevNode:
            self._prev_node = EtcdKeysResult(None, node=prevNode)
            """
            #fix this bug
            r = c.write('/foo', None, dir=True, ttl=50)
            print(r.dir) #True
            r2 = c.write('/foo', None, dir=True, ttl=120, prevExist=True)
            print(r2.dir) #False
            """
            if self._prev_node.dir and not self.dir:
                self.dir = True

    def parse_response(self, response):

        if response.status == httplib.CREATED:
            self.newKey = True

        headers = response.headers
        self.etcd_index = int(headers.get('x-etcd-index', 1))
        self.raft_index = int(headers.get('x-raft-index', 1))
        self.raft_term = int(headers.get('x-raft-term', 0))

    def get_subtree(self, leaves_only=False):

        if not self._children:
            yield self
            return

        if not leaves_only:
            yield self

        for n in self._children:
            node = EtcdKeysResult(None, n)
            for child in node.get_subtree(leaves_only=leaves_only):
                yield child

        return

    @property
    def leaves(self):
        return self.get_subtree(leaves_only=True)

    def __eq__(self, other):

        if not isinstance(other, EtcdKeysResult):
            return False

        for k in self._node_props.keys():
            try:
                a = getattr(self, k)
                b = getattr(other, k)
                if a != b:
                    return False
            except:
                return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return "%s(%r)" % (self.__class__, self.__dict__)


class Response(object):

    REDIRECT_STATUSES = (301, 302, 303, 307, 308)

    def __init__(self, conn=None, status=0, version=0,
                 reason=None, headers=None, body=''):

        self._conn = conn
        self.status = status
        self.version = version
        self.reason = reason
        self.headers = headers
        self._body = body

    @property
    def data(self):

        if self._body:
            return self._body

        if hasattr(self._conn, 'read'):
            self._body = self._conn.read()
            return self._body

        return ''

    def get_redirect_location(self):

        if self.status in self.REDIRECT_STATUSES:
            return self.headers.get('location')

        return None

    @classmethod
    def from_http(Cls, h, **argkv):

        return Cls(h,
                   status=h.status,
                   headers=h.headers,
                   body=h.read_body(None),
                   **argkv)


class Client(object):

    _MGET = 'GET'
    _MPUT = 'PUT'
    _MPOST = 'POST'
    _MDELETE = 'DELETE'

    _write_conditions = set(('prevValue', 'prevIndex', 'prevExist'))
    _read_options = set(('recursive', 'wait', 'waitIndex', 'sorted', 'quorum'))
    _del_conditions = set(('prevValue', 'prevIndex'))

    def __init__(self,
                 host='127.0.0.1',
                 port=2379,
                 version_prefix='/v2',
                 read_timeout=10,
                 allow_redirect=True,
                 protocol='http',
                 allow_reconnect=True,
                 basic_auth_account=None,
                 ):

        self._protocol = protocol
        if protocol == 'https':
            raise EtcdSSLError('not supported https right now')

        self._machines_cache = []
        if not list_type(host):
            self._host = host
            self._port = int(port)
            self._base_uri = '%s://%s:%d' % (self._protocol,
                                             self._host, self._port)
        else:
            for h in host:
                if list_type(h):
                    _h, _p = (list(h) + [int(port)])[:2]
                else:
                    _h, _p = h, int(port)
                self._machines_cache.append('%s://%s:%d' % (self._protocol,
                                                            _h, _p))

            self._base_uri = self._machines_cache.pop(0)
            _, self._host, self._port = self._extract_base_uri()

        self.version_prefix = version_prefix
        self._keys_path = self.version_prefix + '/keys'
        self._stats_path = self.version_prefix + '/stats'
        self._mem_path = self.version_prefix + '/members'
        self._user_path = self.version_prefix + '/auth/users'
        self._role_path = self.version_prefix + '/auth/roles'
        self._read_timeout = read_timeout
        self._allow_redirect = allow_redirect
        self._allow_reconnect = allow_reconnect
        self.basic_auth_account = basic_auth_account

        if self._allow_reconnect:
            if len(self._machines_cache) <= 0:
                self._machines_cache = self.machines
            if self._base_uri in self._machines_cache:
                self._machines_cache.remove(self._base_uri)
        else:
            self._machines_cache = []

    @property
    def base_uri(self):
        return self._base_uri

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port

    @property
    def protocol(self):
        return self._protocol

    @property
    def read_timeout(self):
        return self._read_timeout

    @property
    def allow_redirect(self):
        return self._allow_redirect

    @property
    def machines(self):
        res = self.api_execute(self.version_prefix + '/machines',
                               self._MGET,
                               need_refresh_machines=False)

        nodes = res.data.split(',')

        return [n.strip() for n in nodes]

    @property
    def members(self):
        res = self.api_execute(self._mem_path, self._MGET)

        return utfjson.load(res.data)['members']

    @property
    def leader(self):
        res = self.api_execute(self._stats_path + '/self', self._MGET)
        self_st = utfjson.load(res.data)

        leader_id = self_st.get('leaderInfo', {}).get('leader')
        if leader_id is None:
            return None

        mems = self.members
        for mem in mems:
            if mem['id'] != leader_id:
                continue

            return mem.copy()

    @property
    def version(self):
        res = self.api_execute('/version', self._MGET)

        return utfjson.load(res.data)

    @property
    def st_leader(self):
        leader = self.leader
        if leader is None:
            return None

        leaderhosts = []
        for url in leader['clientURLs']:
            if not url.startswith(self._protocol):
                url = self._protocol + '://' + url
            p = urlparse.urlparse(url)
            if p.hostname == '127.0.0.1':
                continue

            port = p.port or self.port
            leaderhosts.append((p.hostname, port))

        return Client(host=leaderhosts)._st('/leader')

    @property
    def st_self(self):
        return self._st('/self')

    @property
    def st_store(self):
        return self._st('/store')

    @property
    def names(self):
        return [n['name'] for n in self.members]

    @property
    def ids(self):
        return [n['id'] for n in self.members]

    @property
    def clienturls(self):
        return sum([n['clientURLs'] for n in self.members], [])

    @property
    def peerurls(self):
        return sum([n['peerURLs'] for n in self.members], [])

    def __contains__(self, key):

        try:
            self.get(key)
            return True
        except EcodeKeyNotFound:
            return False

    def _sanitize_key(self, key):
        if not key.startswith('/'):
            key = "/{key}".format(key=key)
        return key

    def _extract_base_uri(self):
        p = urlparse.urlparse(self._base_uri)
        return p.scheme, p.hostname, p.port

    def _parse_url(self, url):

        p = urlparse.urlparse(url)

        if p.scheme == 'https':
            raise EtcdSSLError('not supported https right now. ' + url)
        elif p.scheme != 'http':
            return None, None, url

        port = p.port or self.port

        return p.hostname, port, p.path

    def _generate_params(self, options, argkv):

        params = {}
        for k, v in argkv.items():
            if k not in options:
                continue

            if isinstance(v, bool):
                params[k] = v and "true" or "false"
                continue

            params[k] = v

        return params

    def _st(self, st_path):
        st_path = self._sanitize_key(st_path)
        response = self.api_execute(self._stats_path + st_path, self._MGET)
        return self._to_dict(response)

    def _to_keysresult(self, response):

        try:
            res = utfjson.load(response.data)
            r = EtcdKeysResult(**res)
            r.parse_response(response)
            return r
        except ValueError as e:
            logger.error(repr(e) + ' while decode {data}'.format(
                                   data=response.data))
            raise EtcdIncompleteRead('failed to decode %s' % response.data)
        except Exception as e:
            logger.error(repr(e) + ' while decode {data}'.format(
                                   data=response.data))
            raise EtcdResponseError('failed to decode %s' % response.data)

    def _to_dict(self, response):

        try:
            return utfjson.load(response.data)
        except ValueError as e:
            logger.error(repr(e) + ' while decode {data}'.format(
                                   data=response.data))
            raise EtcdIncompleteRead('failed to decode %s' % response.data)
        except Exception as e:
            logger.error(repr(e) + ' while decode {data}'.format(
                                   data=response.data))
            raise EtcdResponseError('failed to decode %s' % response.data)

    def _handle_server_response(self, response):

        if response.status in (httplib.OK, httplib.CREATED,
                               httplib.NO_CONTENT):
            return response

        logger.debug('invalid response status:{st} body:{body}'.format(
                     st=response.status, body=response.data))

        EtcdError.handle(response)

    def _request(self, url, method, params, timeout, bodyinjson):

        while True:
            host, port, path = self._parse_url(url)
            if host is None or port is None or path is None:
                raise EtcdException('url is invalid, {url}'.format(url=url))

            qs = {}
            headers = {}
            body = ''

            if method in (self._MGET, self._MDELETE):
                qs.update(params or {})

                # use once, coz params is in location's query string
                params = None
                headers['Content-Length'] = 0

            elif method in (self._MPUT, self._MPOST):
                if bodyinjson:
                    if params is not None:
                        body = utfjson.dump(params)
                    headers.update({'Content-Type': 'application/json',
                                    'Content-Length': len(body)})
                else:
                    body = urllib.urlencode(params or {})
                    headers.update(
                        {'Content-Type': 'application/x-www-form-urlencoded',
                         'Content-Length': len(body)}
                    )
            else:
                raise EtcdRequestError('HTTP method {method} not supported'
                                       ''.format(method=method))

            if len(qs) > 0:
                if '?' in path:
                    path = path + '&' + urllib.urlencode(qs)
                else:
                    path = path + '?' + urllib.urlencode(qs)

            if self.basic_auth_account is not None:
                auth = {
                    'Authorization': 'Basic {ant}'.format(
                        ant=self.basic_auth_account.encode('base64').strip()),
                }
                headers.update(auth)

            logger.debug('connect -> {mtd} {url}{path} {timeout}'.format(
                         mtd=method,
                         url=self._base_uri,
                         path=path,
                         timeout=timeout))

            h = http.Client(host, port, timeout)
            h.send_request(path, method, headers)
            h.send_body(body)
            h.read_response()

            resp = Response.from_http(h)

            if not self.allow_redirect:
                return resp

            if resp.status not in Response.REDIRECT_STATUSES:
                return resp

            url = resp.get_redirect_location()
            if url is None:
                raise EtcdResponseError('location not found in {header}'
                                        ''.format(header=resp.headers))

            logger.debug('redirect -> ' + url)

    def _api_execute_with_retry(self,
                                path,
                                method,
                                params=None,
                                timeout=None,
                                bodyinjson=False,
                                raise_read_timeout=False,
                                **request_kw):

        # including _base_uri, there are len(_machines_cache) + 1 hosts to try
        # to connect to.
        for i in range(len(self._machines_cache) + 1):

            url = self._base_uri + path

            try:
                response = self._request(url, method, params,
                                         timeout, bodyinjson)
                break
            except (socket.error,
                    http.HttpError) as e:

                if raise_read_timeout and isinstance(e, socket.timeout):
                    raise EtcdReadTimeoutError(e)

                if len(self._machines_cache) > 0:
                    self._machines_cache.append(self._base_uri)
                    self._base_uri = self._machines_cache.pop(0)
                    self._protocol, self._host, self._port = self._extract_base_uri()

                    logger.info('{err} while connect {cur}, try connect {nxt}'
                                .format(err=repr(e), cur=url,
                                        nxt=self._base_uri))

                else:
                    logger.info('no more host to retry')

            except Exception as e:
                logger.exception(repr(e) + ' while send request to etcd')
                raise EtcdException(e)

        else:
            raise NoMoreMachineError('No more machines in the cluster')

        return self._handle_server_response(response)

    def api_execute(self,
                    path,
                    method,
                    params=None,
                    timeout=None,
                    bodyinjson=False,
                    raise_read_timeout=False,
                    need_refresh_machines=True,
                    **request_kw):

        if timeout is None:
            timeout = self.read_timeout
        if timeout == 0:
            timeout = None

        if not path.startswith('/'):
            raise ValueError('Path does not start with /')

        for i in range(0, 2):

            try:

                return self._api_execute_with_retry(
                    path,
                    method,
                    params=params,
                    timeout=timeout,
                    bodyinjson=bodyinjson,
                    raise_read_timeout=raise_read_timeout,
                    **request_kw)

            except NoMoreMachineError as e:

                logger.info(repr(e) + ' while send_request path:{path}, '
                            'method:{mtd}'.format(path=path, mtd=method))

                if i == 1 or not need_refresh_machines or not self._allow_reconnect:
                    raise

                new_machines = self.machines
                old_machines = self._machines_cache + [self._base_uri]

                if set(new_machines) == set(old_machines):
                    raise

                self._machines_cache = new_machines
                self._base_uri = self._machines_cache.pop(0)
                self._protocol, self._host, self._port = self._extract_base_uri()

    def read(self, key, **argkv):

        key = self._sanitize_key(key)

        params = self._generate_params(self._read_options, argkv)

        timeout = argkv.get('timeout')

        response = self.api_execute(self._keys_path + key, self._MGET,
                                    params=params, timeout=timeout)

        return self._to_keysresult(response)

    get = read

    def write(self, key, value=None, ttl=None,
              dir=False, append=False, refresh=False, **argkv):

        key = self._sanitize_key(key)

        params = {}
        if ttl is not None:
            params['ttl'] = ttl

        if dir and value is not None:
            raise EtcdRequestError(
                'Cannot create a directory with a value ' + repr(value))
        elif value is not None:
            params['value'] = value
        elif dir:
            params['dir'] = "true"

        if refresh:
            params['refresh'] = "true"

        params.update(self._generate_params(self._write_conditions, argkv))

        method = append and self._MPOST or self._MPUT
        if '_endpoint' in argkv:
            path = argkv['_endpoint'] + key
        else:
            path = self._keys_path + key

        response = self.api_execute(path, method, params=params)

        return self._to_keysresult(response)

    def test_and_set(self, key, value, ttl=None, **argkv):
        return self.write(key, value=value, ttl=ttl, **argkv)

    def set(self, key, value, ttl=None):
        return self.write(key, value=value, ttl=ttl)

    def update(self, res):

        argkv = {
            'dir': res.dir,
            'ttl': res.ttl,
            'prevExist': True,
        }

        if not res.dir:
            # prevIndex on a dir causes a 'not a file' error. d'oh!
            argkv['prevIndex'] = res.modifiedIndex

        return self.write(res.key, value=res.value, **argkv)

    def delete(self, key, recursive=None, dir=None, **argkv):

        key = self._sanitize_key(key)

        params = {}
        if recursive is not None:
            params['recursive'] = recursive and "true" or "false"

        if dir is not None:
            params['dir'] = dir and "true" or "false"

        params.update(self._generate_params(self._del_conditions, argkv))

        response = self.api_execute(self._keys_path + key, self._MDELETE,
                                    params=params)

        return self._to_keysresult(response)

    def test_and_delete(self, key, **argkv):
        return self.delete(key, **argkv)

    def watch(self, key, waitindex=None, timeout=None, **argkv):

        newest_v = self.get(key)

        if waitindex is None:
            waitindex = argkv.get('waitIndex')

        if waitindex is not None and 0 < waitindex <= newest_v.modifiedIndex:
            return newest_v
        else:
            waitindex = newest_v.etcd_index + 1
            return self._watch(key, waitindex, timeout, **argkv)

    def _watch(self, key, waitindex=None, timeout=None, **argkv):

        key = self._sanitize_key(key)

        params = self._generate_params(self._read_options, argkv)
        params['wait'] = 'true'

        if waitindex is not None:
            params['waitIndex'] = waitindex

        # timeout is 0 means infinite waiting
        while True and timeout == 0:
            try:
                response = self.api_execute(self._keys_path + key,
                                            self._MGET, params=params,
                                            timeout=timeout)
                return self._to_keysresult(response)
            except EtcdIncompleteRead:
                pass

        timeout = timeout or self.read_timeout

        while True:
            st = time.time()
            try:
                response = self.api_execute(self._keys_path + key, self._MGET,
                                            params=params, timeout=timeout,
                                            raise_read_timeout=True)
                return self._to_keysresult(response)
            except (EtcdIncompleteRead, EtcdReadTimeoutError):
                timeout = timeout - (time.time() - st)
                if timeout <= 0:
                    raise EtcdReadTimeoutError('Watch Timeout: ' + key)

    def eternal_watch(self, key, waitindex=None, until=None, **argkv):

        local_index = waitindex
        while True:
            res = self._watch(key, waitindex=local_index, timeout=0, **argkv)
            if until is not None and res.modifiedIndex is not None:
                if res.modifiedIndex >= until:
                    yield res
                    return

            if local_index is not None:
                local_index = (res.modifiedIndex or local_index) + 1

            yield res

    def mkdir(self, key, ttl=None, **argkv):
        return self.write(key, ttl=ttl, dir=True, **argkv)

    def refresh(self, key, ttl=None, **argkv):
        argkv['prevExist'] = True
        return self.write(key, ttl=ttl, refresh=True, **argkv)

    def lsdir(self, key, **argkv):
        return self.read(key, **argkv)

    def rlsdir(self, key, **argkv):
        argkv['recursive'] = True
        return self.read(key, **argkv)

    def deldir(self, key, **argkv):
        return self.delete(key, dir=True, **argkv)

    def rdeldir(self, key, **argkv):
        argkv['recursive'] = True
        return self.delete(key, dir=True, **argkv)

    def add_member(self, *peerurls):
        if len(peerurls) == 0:
            raise EtcdException('no peer url found')

        data = {'peerURLs': peerurls}
        response = self.api_execute(self._mem_path, self._MPOST,
                                    params=data, bodyinjson=True)

        return self._to_dict(response)

    def del_member(self, mid):
        if mid not in self.ids:
            logger.info('{mid} not in the cluster when delete member'.format(
                        mid=mid))
            return

        mid = self._sanitize_key(mid)
        self.api_execute(self._mem_path + mid, self._MDELETE)

    def change_peerurls(self, mid, *peerurls):
        if mid not in self.ids:
            logger.info('{mid} not in the cluster when change peerurls'.format(
                        mid=mid))
            return

        if len(peerurls) == 0:
            raise EtcdException('no peer url found')

        mid = self._sanitize_key(mid)
        data = {'peerURLs': peerurls}

        self.api_execute(self._mem_path + mid, self._MPUT,
                         params=data, bodyinjson=True)

    def _root_auth(self, password):
        return 'root:%s' % (password)

    def create_root(self, password):

        path = self._user_path + '/root'
        params = {'user': 'root', 'password': password}

        res = self.api_execute(path, self._MPUT,
                               params=params, bodyinjson=True)
        return self._to_dict(res)

    def enable_auth(self, root_password):
        self.basic_auth_account = self._root_auth(root_password)
        self.api_execute('/v2/auth/enable', self._MPUT)

    def disable_auth(self, root_password):
        self.basic_auth_account = self._root_auth(root_password)
        self.api_execute('/v2/auth/enable', self._MDELETE)

    def create_user(self, name, password, root_password, roles=None):

        self.basic_auth_account = self._root_auth(root_password)
        path = self._user_path + self._sanitize_key(name)
        if roles is not None:
            params = {"user": name, "password": password, "roles": roles}
        else:
            params = {"user": name, "password": password}

        res = self.api_execute(
            path, self._MPUT, params=params, bodyinjson=True)
        return self._to_dict(res)

    def create_role(self, name, root_password, permissions=None):

        self.basic_auth_account = self._root_auth(root_password)
        path = self._role_path + self._sanitize_key(name)
        if permissions is not None:
            params = {"role": name, "permissions": {"kv": permissions}}
        else:
            params = {"role": name}

        res = self.api_execute(path, self._MPUT,
                               params=params, bodyinjson=True)
        return self._to_dict(res)

    def get_user(self, name, root_password):

        self.basic_auth_account = self._root_auth(root_password)
        if name is not None:
            path = self._user_path + self._sanitize_key(name)
        else:
            path = self._user_path

        res = self.api_execute(path, self._MGET)
        return self._to_dict(res)

    def get_role(self, name, root_password):

        self.basic_auth_account = self._root_auth(root_password)
        if name is not None:
            path = self._role_path + self._sanitize_key(name)
        else:
            path = self._role_path

        res = self.api_execute(path, self._MGET)
        return self._to_dict(res)

    def grant_user_roles(self, name, root_password, roles):

        self.basic_auth_account = self._root_auth(root_password)
        path = self._user_path + self._sanitize_key(name)
        params = {'user': name, 'grant': roles}

        res = self.api_execute(path, self._MPUT,
                               params=params, bodyinjson=True)
        return self._to_dict(res)

    def revoke_user_roles(self, name, root_password, roles):

        self.basic_auth_account = self._root_auth(root_password)
        path = self._user_path + self._sanitize_key(name)
        params = {'user': name, 'revoke': roles}

        res = self.api_execute(path, self._MPUT,
                               params=params, bodyinjson=True)
        return self._to_dict(res)

    def grant_role_permissions(self, name, root_password, permissions):

        self.basic_auth_account = self._root_auth(root_password)
        path = self._role_path + self._sanitize_key(name)
        params = {"role": name, "grant": {"kv": permissions}}

        res = self.api_execute(path, self._MPUT,
                               params=params, bodyinjson=True)
        return self._to_dict(res)

    def revoke_role_permissions(self, name, root_password, permissions):

        self.basic_auth_account = self._root_auth(root_password)
        path = self._role_path + self._sanitize_key(name)
        params = {"role": name, "revoke": {"kv": permissions}}

        res = self.api_execute(path, self._MPUT,
                               params=params, bodyinjson=True)
        return self._to_dict(res)

    def delete_user(self, user_name, root_password):

        self.basic_auth_account = self._root_auth(root_password)
        path = self._user_path + self._sanitize_key(user_name)
        self.api_execute(path, self._MDELETE)

    def delete_role(self, role_name, root_password):

        self.basic_auth_account = self._root_auth(root_password)
        path = self._role_path + self._sanitize_key(role_name)
        self.api_execute(path, self._MDELETE)



#!/usr/bin/env python2
# coding: utf-8

import ctypes
import inspect
import threading
import time


class AsyncRaiseError(Exception):
    pass


class InvalidThreadIdError(AsyncRaiseError):
    pass


def start_thread(target, name=None, args=None, kwargs=None, daemon=False, after=None):
    args = args or ()
    kwargs = kwargs or {}

    if after is None:
        _target = target
    else:
        def _target(*args, **kwargs):
            time.sleep(after)
            target(*args, **kwargs)

    t = threading.Thread(target=_target, name=name, args=args, kwargs=kwargs)
    t.daemon = daemon
    t.start()

    return t


def start_daemon(target, name=None, args=None, kwargs=None, after=None):
    return start_thread(target, name=name, args=args, kwargs=kwargs,
                        daemon=True, after=after)


def start_daemon_thread(target, name=None, args=None, kwargs=None):
    # deprecated
    return start_daemon(target, name=name, args=args, kwargs=kwargs)


def raise_in_thread(thread, exctype):
    """
    Asynchronously raises an exception in the context of the given thread,
    which should cause the thread to exit silently (unless caught).

    Raises:
        - InvalidThreadIdError if the given thread is not alive or found.
        - TypeError or ValueError if any input is illegal.
        - AsyncRaiseError for other unexpected errors.
    """
    if not isinstance(thread, threading.Thread):
        raise TypeError("Only Thread is allowed, got {t}".format(t=thread))

    _async_raise(thread.ident, exctype)


def _async_raise(tid, exctype):
    """raises the exception, performs cleanup if needed"""
    if not inspect.isclass(exctype):
        raise TypeError("Only types can be raised (not instances)")

    if not issubclass(exctype, BaseException):
        raise ValueError("Only sub classes of BaseException can be raised")

    # PyThreadState_SetAsyncExc requires GIL to be held
    gil_state = ctypes.pythonapi.PyGILState_Ensure()
    try:
        res = ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(tid),
                                                         ctypes.py_object(exctype))
        if res == 0:
            # The thread is likely dead already.
            raise InvalidThreadIdError(tid)

        elif res != 1:
            # If more than one threads are affected (WTF?), we're in trouble, and
            # we try our best to revert the effect, although this may not work.
            ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(tid), None)

            raise AsyncRaiseError("PyThreadState_SetAsyncExc failed", res)

    finally:
        ctypes.pythonapi.PyGILState_Release(gil_state)



#!/usr/bin/env python
# coding: utf-8

import logging

from kazoo.exceptions import BadVersionError

from pykit import rangeset
from pykit import utfjson
from pykit import zkutil

from .status import STATUS
from .storage import Storage
from .zkaccessor import ZKKeyValue
from .zkaccessor import ZKValue

logger = logging.getLogger(__name__)


class ZKStorage(Storage):

    conflicterror = BadVersionError

    def __init__(self, zke):

        assert isinstance(zke, zkutil.KazooClientExt)

        self.zke = zke

        self.txidset = ZKValue(self.zke,
                               self.zke._zkconf.txidset,
                               load=txidset_load)

        self.journal = ZKKeyValue(self.zke,
                                  self.zke._zkconf.journal)

        self.record = ZKKeyValue(self.zke,
                                 self.zke._zkconf.record,
                                 nonode_callback=record_nonode_cb)

        self.state = ZKKeyValue(self.zke,
                                self.zke._zkconf.tx_state,
                                nonode_callback=state_nonode_cb,
                                )

    def acquire_key_loop(self, txid, key, timeout):

        logger.info('watch acquire: {txid} {key}'.format(txid=txid, key=key))

        keylock = self._make_key_lock(txid, key)

        for holder, ver in keylock.acquire_loop(timeout=timeout):
            yield int(holder), ver

    def try_release_key(self, txid, key):

        logger.info('releasing: {txid} {key}'.format(txid=txid, key=key))

        keylock = self._make_key_lock(txid, key)

        locked, txid, ver = keylock.try_acquire()
        if locked:
            keylock.release()
        else:
            keylock.close()

        return locked, utfjson.load(txid), ver

    def _make_key_lock(self, txid, key):
        keylock = zkutil.ZKLock(key,
                                zkclient=self.zke,
                                zkconf=self.zke._zkconf,
                                ephemeral=False,
                                identifier=utfjson.dump(txid))

        return keylock


def txidset_load(value):
    rst = {}
    for k in STATUS:
        if k not in value:
            value[k] = []
        rst[k] = rangeset.RangeSet(value[k])
    return rst


def record_nonode_cb():
    """
    If NoNodeError received, make a default value for a record
    """

    return [(-1, None)], -1


def state_nonode_cb():
    return None, -1



#!/usr/bin/env python
# coding: utf-8

import logging
import unittest

from pykit import mysqlutil
from pykit import ututil

dd = ututil.dd

logger = logging.getLogger(__name__)


class TestPrivileges(unittest.TestCase):

    def test_load_dump(self):
        ks = (
            "ALL",
            "ALTER",
            "ALTER ROUTINE",
            "CREATE",
            "CREATE ROUTINE",
            "CREATE TABLESPACE",
            "CREATE TEMPORARY TABLES",
            "CREATE USER",
            "CREATE VIEW",
            "DELETE",
            "DROP",
            "EVENT",
            "EXECUTE",
            "FILE",
            "GRANT OPTION",
            "INDEX",
            "INSERT",
            "LOCK TABLES",
            "PROCESS",
            "PROXY",
            "REFERENCES",
            "RELOAD",
            "REPLICATION CLIENT",
            "REPLICATION SLAVE",
            "SELECT",
            "SHOW DATABASES",
            "SHOW VIEW",
            "SHUTDOWN",
            "SUPER",
            "TRIGGER",
            "UPDATE",
            "USAGE",
        )

        for k in ks:
            self.assertEqual((k,), mysqlutil.privileges[k])
            self.assertEqual((k,), mysqlutil.privileges[k.replace(' ', '_')])

        shortcuts = {
            'replicator': (
                'REPLICATION CLIENT',
                'REPLICATION SLAVE',
                'SELECT',
            ),
            'monitor': (
                'SELECT',
            ),
            'business': (
                'CREATE',
                'DROP',
                'REFERENCES',
                'ALTER',
                'DELETE',
                'INDEX',
                'INSERT',
                'SELECT',
                'UPDATE',
            ),
            'readwrite': (
                'DELETE',
                'INSERT',
                'SELECT',
                'UPDATE',
            ),
        }

        for k, expected in shortcuts.items():
            self.assertEqual(expected, mysqlutil.privileges[k])
